# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the Blinn-Phong specular reflectance.

For a derivation of the normalization factor ensuring energy conservation, we
refer the interested reader to:
Fabian Giesen.
"Derivation of Phong and Blinn-Phong BRDF normalization factors". 2009
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math

import tensorflow as tf

from tensorflow_graphics.math import vector
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import shape


def _brdf_normalization_factor(shininess):
  """Returns the normalization factor needed to ensure energy conservation."""
  numerator = (shininess + 2.0) * (shininess + 4.0)
  denominator = 8.0 * math.pi * (
      tf.pow(tf.constant(2.0, dtype=shininess.dtype), -shininess / 2.0) +
      shininess)
  return safe_ops.safe_signed_div(numerator, denominator)


def brdf(direction_incoming_light,
         direction_outgoing_light,
         surface_normal,
         shininess,
         albedo,
         brdf_normalization=True,
         name=None):
  """Evaluates the specular brdf of the Blinn-Phong model.

  Note:
    In the following, A1 to An are optional batch dimensions, which must be
    broadcast compatible.

  Note:
    The gradient of this function is not smooth when the dot product of the
    normal with any light is 0.0.

  Args:
    direction_incoming_light: A tensor of shape `[A1, ..., An, 3]`, where the
      last dimension represents a normalized incoming light vector.
    direction_outgoing_light: A tensor of shape `[A1, ..., An, 3]`, where the
      last dimension represents a normalized outgoing light vector.
    surface_normal: A tensor of shape `[A1, ..., An, 3]`, where the last
      dimension represents a normalized surface normal.
    shininess: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      represents a non-negative shininess coefficient.
    albedo: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents albedo with values in [0,1].
    brdf_normalization: A `bool` indicating whether normalization should be
      applied to enforce the energy conservation property of BRDFs. Note that
      `brdf_normalization` must be set to False in order to use the original
      Blinn-Phong specular model.
    name: A name for this op. Defaults to "blinn_phong_brdf".

  Returns:
      A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
        the amount of light reflected in the outgoing light direction.

  Raises:
    ValueError: if the shape of `direction_incoming_light`,
    `direction_outgoing_light`, `surface_normal`, `shininess` or `albedo` is not
    supported.
    InvalidArgumentError: if not all of shininess values are non-negative, or if
    at least one element of `albedo` is outside of [0,1].
  """
  with tf.compat.v1.name_scope(name, "blinn_phong_brdf", [
      direction_incoming_light, direction_outgoing_light, surface_normal,
      shininess, albedo
  ]):
    direction_incoming_light = tf.convert_to_tensor(
        value=direction_incoming_light)
    direction_outgoing_light = tf.convert_to_tensor(
        value=direction_outgoing_light)
    surface_normal = tf.convert_to_tensor(value=surface_normal)
    shininess = tf.convert_to_tensor(value=shininess)
    albedo = tf.convert_to_tensor(value=albedo)

    shape.check_static(
        tensor=direction_incoming_light,
        tensor_name="direction_incoming_light",
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=direction_outgoing_light,
        tensor_name="direction_outgoing_light",
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=surface_normal,
        tensor_name="surface_normal",
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=shininess, tensor_name="shininess", has_dim_equals=(-1, 1))
    shape.check_static(
        tensor=albedo, tensor_name="albedo", has_dim_equals=(-1, 3))
    shape.compare_batch_dimensions(
        tensors=(direction_incoming_light, direction_outgoing_light,
                 surface_normal, shininess, albedo),
        tensor_names=("direction_incoming_light", "direction_outgoing_light",
                      "surface_normal", "shininess", "albedo"),
        last_axes=-2,
        broadcast_compatible=True)
    direction_incoming_light = asserts.assert_normalized(
        direction_incoming_light)
    direction_outgoing_light = asserts.assert_normalized(
        direction_outgoing_light)
    surface_normal = asserts.assert_normalized(surface_normal)
    albedo = asserts.assert_all_in_range(albedo, 0.0, 1.0, open_bounds=False)
    shininess = asserts.assert_all_above(shininess, 0.0, open_bound=False)

    # Checks whether the incoming or outgoing light point behind the surface.
    dot_incoming_light_surface_normal = vector.dot(-direction_incoming_light,
                                                   surface_normal)
    dot_outgoing_light_surface_normal = vector.dot(direction_outgoing_light,
                                                   surface_normal)
    min_dot = tf.minimum(dot_incoming_light_surface_normal,
                         dot_outgoing_light_surface_normal)
    difference_outgoing_incoming = (
        direction_outgoing_light - direction_incoming_light)
    difference_outgoing_incoming = tf.math.l2_normalize(
        difference_outgoing_incoming, axis=-1)
    cos_alpha = vector.dot(
        surface_normal, difference_outgoing_incoming, axis=-1)
    cos_alpha = tf.maximum(cos_alpha, tf.zeros_like(cos_alpha))
    blinn_phong_model = albedo * tf.pow(cos_alpha, shininess)
    if brdf_normalization:
      blinn_phong_model *= _brdf_normalization_factor(shininess)
    common_shape = shape.get_broadcasted_shape(min_dot.shape,
                                               blinn_phong_model.shape)
    d_val = lambda dim: 1 if dim is None else tf.compat.v1.dimension_value(dim)
    common_shape = [d_val(dim) for dim in common_shape]
    condition = tf.broadcast_to(tf.greater_equal(min_dot, 0.0), common_shape)
    blinn_phong_model = tf.broadcast_to(blinn_phong_model, common_shape)
    return tf.compat.v1.where(condition, blinn_phong_model,
                              tf.zeros_like(blinn_phong_model))


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements TensorFlow dual quaternion utility functions.

A dual quaternion is an extension of a quaternion with the real and dual parts
and written as $$q = q_r + epsilon q_d$$, where $$epsilon$$ is the dual number
with the property $$e^2 = 0$$. It can thus be represented as two quaternions,
and thus stored as 8 numbers. We define the operations in terms of the two
quaternions $$q_r$$ and $$q_d$$.

Dual quaternions are extensions of quaternions to represent rigid
transformations (rotations and translations). They are in particular important
for deforming geometries as linear blending is a very close approximation of
closest path blending, which is not the case for any other representation.

Note: Some of the functions expect normalized quaternions as inputs where
$$|q_r| = 1$$.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow.compat.v2 as tf

from tensorflow_graphics.geometry.transformation import quaternion
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import shape


def conjugate(dual_quaternion, name=None):
  """Computes the conjugate of a dual quaternion.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    dual_quaternion: A tensor of shape `[A1, ..., An, 8]`, where the last
    dimension represents a normalized dual quaternion.
    name: A name for this op that defaults to "dual_quaternion_conjugate".

  Returns:
    A tensor of shape `[A1, ..., An, 8]`, where the last dimension represents
    a normalized dual quaternion.

  Raises:
    ValueError: If the shape of `dual_quaternion` is not supported.
  """
  with tf.compat.v1.name_scope(name, "dual_quaternion_conjugate",
                               [dual_quaternion]):
    dual_quaternion = tf.convert_to_tensor(value=dual_quaternion)

    shape.check_static(
        tensor=dual_quaternion, tensor_name="dual_quaternion",
        has_dim_equals=(-1, 8))

    quaternion_real, quaternion_dual = tf.split(
        dual_quaternion, (4, 4), axis=-1)

    quaternion_real = asserts.assert_normalized(quaternion_real)

    return tf.concat((quaternion.conjugate(quaternion_real),
                      quaternion.conjugate(quaternion_dual)),
                     axis=-1)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for shape utility functions."""

import unittest

from absl.testing import parameterized
import tensorflow as tf

from tensorflow_graphics.util import test_case


class TestCaseTest(test_case.TestCase):

  def _dummy_tf_lite_compatible_function(self, data):
    """Executes a simple supported function to test TFLite conversion."""
    data = tf.convert_to_tensor(value=data)
    return 2.0 * data

  def _dummy_tf_lite_incompatible_function(self, data):
    """Executes a simple unsupported function to test TFLite conversion."""
    del data  # Unused
    return 2.0 * tf.ones(shape=[2] * 10)

  @parameterized.parameters(None, (((1.0,),),))
  def test_assert_tf_lite_convertible_exception_not_raised(self, test_inputs):
    """Tests that assert_tf_lite_convertible succeeds with a simple function."""
    tc = test_case.TestCase(methodName="assert_tf_lite_convertible")

    # We can't use self.assert_exception_is_not_raised here because we need to
    # use `shapes` as both a named argument and a kwarg.
    try:
      tc.assert_tf_lite_convertible(
          func=self._dummy_tf_lite_compatible_function,
          shapes=((1,),),
          test_inputs=test_inputs)
    except unittest.SkipTest as e:
      # Forwarding SkipTest exception in order to skip the test.
      raise e
    except Exception as e:  # pylint: disable=broad-except
      self.fail("Exception raised: %s" % type(e))

  @parameterized.parameters(None, (((1.0,),),))
  def test_assert_tf_lite_convertible_exception_raised(self, test_inputs):
    """Tests that assert_tf_lite_convertible succeeds with a simple function."""
    # TODO(b/131912561): TFLite conversion throws SIGABRT instead of Exception.
    return
    # pylint: disable=unreachable
    # This code should be able to catch exceptions correctly once TFLite bug
    # is fixed.
    tc = test_case.TestCase(methodName="assert_tf_lite_convertible")

    with self.assertRaises(Exception):
      tc.assert_tf_lite_convertible(
          func=self._dummy_tf_lite_incompatible_function,
          shapes=((1,),),
          test_inputs=test_inputs)
    # pylint: enable=unreachable

  def _dummy_failing_function(self, data):
    """Fails instantly."""
    del data  # Unused
    raise ValueError("Fail.")

  def test_assert_exception_is_not_raised_raises_exception(self):
    """Tests that assert_exception_is_not_raised raises exception."""
    if tf.executing_eagerly():
      # In eager mode placeholders are assigned zeros by default, which fails
      # for various tests. Therefore this function can only be tested in graph
      # mode.
      return
    tc = test_case.TestCase(methodName="assert_exception_is_not_raised")

    with self.assertRaises(AssertionError):
      tc.assert_exception_is_not_raised(
          self._dummy_failing_function, shapes=((1,),))


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for quadratic_radial_distortion."""

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.rendering.camera import quadratic_radial_distortion
from tensorflow_graphics.util import test_case

RANDOM_TESTS_NUM_IMAGES = 10
RANDOM_TESTS_HEIGHT = 8
RANDOM_TESTS_WIDTH = 8

RADII_SHAPE = (RANDOM_TESTS_NUM_IMAGES, RANDOM_TESTS_HEIGHT, RANDOM_TESTS_WIDTH)
COEFFICIENT_SHAPE = (RANDOM_TESTS_NUM_IMAGES,)


def _get_random_radii():
  return np.random.rand(*RADII_SHAPE).astype('float32')


def _get_zeros_radii():
  return np.zeros(shape=RADII_SHAPE).astype('float32')


def _get_ones_radii():
  return np.ones(shape=RADII_SHAPE).astype('float32')


def _get_random_coefficient():
  return np.random.rand(*COEFFICIENT_SHAPE).astype('float32')


def _get_zeros_coefficient():
  return np.zeros(shape=COEFFICIENT_SHAPE).astype('float32')


def _get_ones_coefficient():
  return np.ones(shape=COEFFICIENT_SHAPE).astype('float32')


def _make_shape_compatible(coefficients):
  return np.expand_dims(np.expand_dims(coefficients, axis=-1), axis=-1)


class QuadraticRadialDistortionTest(test_case.TestCase):

  def test_distortion_factor_random_positive_distortion_coefficient(self):
    """Tests that distortion_factor produces the expected outputs."""
    squared_radii = _get_random_radii() * 2.0
    distortion_coefficient = _get_random_coefficient() * 2.0

    distortion, mask = quadratic_radial_distortion.distortion_factor(
        squared_radii, distortion_coefficient)

    distortion_coefficient = _make_shape_compatible(distortion_coefficient)
    with self.subTest(name='distortion'):
      self.assertAllClose(1.0 + distortion_coefficient * squared_radii,
                          distortion)

    # No overflow when distortion_coefficient >= 0.0.
    with self.subTest(name='mask'):
      self.assertAllInSet(mask, (False,))

  def test_distortion_factor_preset_zero_distortion_coefficient(self):
    """Tests distortion_factor at zero distortion coefficient."""
    squared_radii = _get_random_radii() * 2.0

    distortion, mask = quadratic_radial_distortion.distortion_factor(
        squared_radii, 0.0)

    with self.subTest(name='distortion'):
      self.assertAllClose(tf.ones_like(squared_radii), distortion)

    # No overflow when distortion_coefficient = 0.0.
    with self.subTest(name='mask'):
      self.assertAllInSet(mask, (False,))

  def test_distortion_factor_random_negative_distortion_coefficient(self):
    """Tests that distortion_factor produces the expected outputs."""
    squared_radii = _get_random_radii() * 2.0
    distortion_coefficient = _get_random_coefficient() * -0.2

    distortion, mask = quadratic_radial_distortion.distortion_factor(
        squared_radii, distortion_coefficient)
    distortion_coefficient = _make_shape_compatible(distortion_coefficient)
    max_squared_radii = -1.0 / 3.0 / distortion_coefficient
    expected_overflow_mask = squared_radii > max_squared_radii
    valid_mask = np.logical_not(expected_overflow_mask)
    # We assert correctness of the mask, and of all the pixels that are not in
    # overflow.
    actual_distortion_when_valid = self.evaluate(distortion)[valid_mask]
    expected_distortion_when_valid = (
        1.0 + distortion_coefficient * squared_radii)[valid_mask]

    with self.subTest(name='distortion'):
      self.assertAllClose(expected_distortion_when_valid,
                          actual_distortion_when_valid)

    with self.subTest(name='mask'):
      self.assertAllEqual(expected_overflow_mask, mask)

  def test_distortion_factor_preset_zero_radius(self):
    """Tests distortion_factor at the corner case of zero radius."""
    squared_radii = _get_zeros_radii()
    distortion_coefficient = _get_random_coefficient() - 0.5

    distortion, mask = quadratic_radial_distortion.distortion_factor(
        squared_radii, distortion_coefficient)

    with self.subTest(name='distortion'):
      self.assertAllClose(np.ones_like(squared_radii), distortion)

    with self.subTest(name='mask'):
      self.assertAllInSet(mask, (False,))

  @parameterized.parameters(quadratic_radial_distortion.distortion_factor,
                            quadratic_radial_distortion.undistortion_factor)
  def test_both_negative_radius_exception_raised(self, distortion_function):
    """Tests that an exception is raised when the squared radius is negative."""
    squared_radii = _get_zeros_radii() - 0.5
    distortion_coefficient = _get_random_coefficient() - 0.5

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(distortion_function(squared_radii, distortion_coefficient))

  @parameterized.parameters((2, 2e-3), (3, 1e-8))
  def test_undistortion_factor_random_positive_distortion_coefficient(
      self, num_iterations, tolerance):
    """Tests that undistortion_factor produces the expected outputs."""
    distorted_squared_radii = _get_random_radii() * 2.0
    distortion_coefficient = _get_random_coefficient() * 0.2

    undistortion, mask = quadratic_radial_distortion.undistortion_factor(
        distorted_squared_radii, distortion_coefficient, num_iterations)
    distortion_coefficient = _make_shape_compatible(distortion_coefficient)
    undistorted_squared_radii = tf.square(
        undistortion) * distorted_squared_radii
    # We distort again the undistorted radii and compare to the original
    # distorted_squared_radii.
    redistorted_squared_radii = tf.square(
        1.0 + distortion_coefficient *
        undistorted_squared_radii) * undistorted_squared_radii

    with self.subTest(name='distortion'):
      self.assertAllClose(
          distorted_squared_radii, redistorted_squared_radii, atol=tolerance)

    # Positive distortion_coefficients never overflow.
    with self.subTest(name='mask'):
      self.assertAllInSet(mask, (False,))

  @parameterized.parameters((2, 2e-2), (3, 6e-3), (4, 6e-4))
  def test_undistortion_factor_random_negative_distortion_coefficient(
      self, num_iterations, tolerance):
    """Tests that undistortion_factor produces the expected outputs."""
    distorted_squared_radii = _get_random_radii() * 2.0
    distortion_coefficient = _get_random_coefficient() * -0.2

    undistortion, mask = quadratic_radial_distortion.undistortion_factor(
        distorted_squared_radii, distortion_coefficient, num_iterations)
    distortion_coefficient = _make_shape_compatible(distortion_coefficient)
    undistorted_squared_radii = tf.square(
        undistortion) * distorted_squared_radii
    # See explanation in the implementation comments for this formula.
    expected_overflow_mask = (
        distorted_squared_radii * distortion_coefficient + 4.0 / 27.0 < 0)
    redistorted_squared_radii = tf.square(
        1.0 + distortion_coefficient *
        undistorted_squared_radii) * undistorted_squared_radii
    valid_mask = np.logical_not(expected_overflow_mask)
    redistorted_squared_radii_when_valid = self.evaluate(
        redistorted_squared_radii)[valid_mask]
    distorted_squared_radii_when_valid = distorted_squared_radii[valid_mask]

    with self.subTest(name='distortion'):
      self.assertAllClose(
          distorted_squared_radii_when_valid,
          redistorted_squared_radii_when_valid,
          rtol=tolerance,
          atol=tolerance)

    # We assert correctness of the mask, and of all the pixels that are not in
    # overflow, distorting again the undistorted radii and comparing to the
    # original distorted_squared_radii.
    with self.subTest(name='mask'):
      self.assertAllEqual(expected_overflow_mask, mask)

  def test_undistortion_factor_zero_distortion_coefficient(self):
    """Tests undistortion_factor at zero distortion coefficient."""
    squared_radii = _get_random_radii() * 2.0

    undistortion, mask = quadratic_radial_distortion.undistortion_factor(
        squared_radii, 0.0)

    with self.subTest(name='distortion'):
      self.assertAllClose(tf.ones_like(squared_radii), undistortion)

    # No overflow when distortion_coefficient = 0.0.
    with self.subTest(name='mask'):
      self.assertAllEqual(np.zeros_like(squared_radii), mask)

  @parameterized.parameters(
      ('must have a rank greater than 1', (2,), (2, 1)),
      ('Not all batch dimensions are broadcast-compatible', (2, 2, 2), (3,)),
      ('Not all batch dimensions are broadcast-compatible', (2, 2, 2), (3, 3)),
  )
  def test_distortion_factor_shape_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(
        func=quadratic_radial_distortion.distortion_factor,
        error_msg=error_msg,
        shapes=shapes)

  @parameterized.parameters(
      ((2, 2), ()),
      ((1, 2, 2), (2,)),
      ((2, 2, 2), (2,)),
      ((2, 2), (2, 2)),
      ((2, 2, 2), (1, 2)),
      ((2, 3, 4), (1,)),
      ((2, 3, 4), (1, 1)),
      ((2, 3, 4), (2,)),
  )
  def test_distortion_factor_shape_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_not_raised(
        func=quadratic_radial_distortion.distortion_factor, shapes=shapes)

  @parameterized.parameters(
      ('must have a rank greater than 1', (2,), (2, 1)),
      ('Not all batch dimensions are broadcast-compatible', (2, 2, 2), (3,)),
      ('Not all batch dimensions are broadcast-compatible', (2, 2, 2), (3, 3)),
  )
  def test_undistortion_factor_shape_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(
        func=quadratic_radial_distortion.undistortion_factor,
        error_msg=error_msg,
        shapes=shapes)

  @parameterized.parameters(
      ((2, 2), ()),
      ((1, 2, 2), (2,)),
      ((2, 2, 2), (2,)),
      ((2, 2), (2, 2)),
      ((2, 2, 2), (1, 2)),
      ((2, 3, 4), (1,)),
      ((2, 3, 4), (1, 1)),
      ((2, 3, 4), (2,)),
  )
  def test_undistortion_factor_shape_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_not_raised(
        func=quadratic_radial_distortion.undistortion_factor, shapes=shapes)

  @parameterized.parameters(quadratic_radial_distortion.distortion_factor,
                            quadratic_radial_distortion.undistortion_factor)
  def test_both_radial_jacobian(self, distortion_function):
    """Test the Jacobians with respect to squared radii."""
    squared_radii = _get_random_radii().astype(np.float64) * 0.5
    distortion_coefficients = _get_random_coefficient().astype(np.float64) * 0.5
    distortion_coefficients -= 0.25

    def distortion_fn(squared_radii):
      distortion, _ = distortion_function(squared_radii,
                                          distortion_coefficients)
      return distortion

    self.assert_jacobian_is_correct_fn(
        distortion_fn, [squared_radii], delta=1e-7, atol=1e-3)

  @parameterized.parameters(quadratic_radial_distortion.distortion_factor,
                            quadratic_radial_distortion.undistortion_factor)
  def test_both_distortion_coefficient_jacobian(self, distortion_function):
    """Test the Jacobians with respect to distortion coefficients."""
    squared_radii = _get_random_radii().astype(np.float64) * 0.5
    distortion_coefficients = _get_random_coefficient().astype(np.float64) * 0.5
    distortion_coefficients -= 0.25

    def distortion_fn(distortion_coefficients):
      distortion, _ = distortion_function(squared_radii,
                                          distortion_coefficients)
      return distortion

    self.assert_jacobian_is_correct_fn(
        distortion_fn, [distortion_coefficients], delta=1e-7, atol=1e-3)


if __name__ == '__main__':
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""pix3d dataset."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os

import numpy as np
import tensorflow as tf
from tensorflow_datasets import features as tfds_features
import tensorflow_datasets.public_api as tfds

from tensorflow_graphics.datasets import features as tfg_features

_CITATION = '''
@inproceedings{pix3d,
  title={Pix3D: Dataset and Methods for Single-Image 3D Shape Modeling},
  author={Sun, Xingyuan and Wu, Jiajun and Zhang, Xiuming and Zhang, Zhoutong
  and Zhang, Chengkai and Xue, Tianfan and Tenenbaum, Joshua B and
  Freeman, William T},
  booktitle={IEEE Conference on Computer Vision and Pattern Recognition (CVPR)},
  year={2018}
}
'''

_DESCRIPTION = '''
Pix3D is a large-scale dataset of diverse image-shape pairs
with pixel-level 2D-3D alignment. It has wide applications in shape-related
 tasks including reconstruction, retrieval, viewpoint estimation, etc.

Pix3D contains 10,069 2D-3D pairs of 395 distinct 3D shapes, categorised into
nine object categories. Each sample comprises of an image, 3D shape represented
as (non-watertight) triangle mesh and voxel grid, bounding-box,
 segmentation mask, intrinsic and extrinsic camera parameters and 2D and 3D key
 points.

Notes:
  * The object and camera poses are provided with respect to the scene, whereas
    the camera is placed at the origin. Pix3D also provides the features
    `camera/position_with_respect_to_object` and `camera/inplane_rotation`.
    Those values are defined in object coordinates and will reproduce an image
    that is equivalent to the original image under a homography transformation.
    They are defined for viewer-centered algorithms whose predictions need to be
    rotated back to the canonical view for evaluations against ground truth
    shapes. This is necessary as most algorithms assume that the camera is
    looking at the object's center, the raw input images are usually cropped or
    transformed before sending into their pipeline.
  * There are two wrong segmentation masks in the annotations of the original
    Pix3D dataset (See https://github.com/xingyuansun/pix3d/issues/18 for
    details). We ignore those samples in this version of the dataset. However,
    if you want to use them, we provide own rendered segmentation masks in
    `tensorflow_graphics/datasets/pix3d/fixed_masks/`. Feel free to copy those
    two masks to your local Pix3D directory in `<PIX3D_HOME>/mask/table/`.
    Additionally, you need to add the indices of these samples to the split
    files located at
    `<TF Graphics Repository>/tensorflow_graphics/datasets/pix3d/splits`.
    The index `7953` needs to be appended to the train index and `9657` belongs
    to the test index.

Train/Test split:
  Pix3D does not provide a standard train/test split. Therefore, this
  implementation adopts the S2 split from Mesh R-CNN
   (https://arxiv.org/abs/1906.02739, Sec. 4.2). This split ensures that the 3D
  models appearing in the train and test sets are disjoint.
'''

_CHECKSUMS_DIR = os.path.normpath(
    os.path.join(os.path.dirname(__file__), 'checksums/')
)
tfds.download.add_checksums_dir(_CHECKSUMS_DIR)


class Pix3d(tfds.core.GeneratorBasedBuilder):
  """Pix3D is a large-scale dataset of diverse image-shape pairs with pixel-level 2D-3D alignment."""

  VERSION = tfds.core.Version('0.1.0')

  TRAIN_SPLIT_IDX = os.path.join(os.path.dirname(__file__),
                                 'splits/pix3d_train.npy')
  TEST_SPLIT_IDX = os.path.join(os.path.dirname(__file__),
                                'splits/pix3d_test.npy')

  CLASS_INDEX = ['background', 'bed', 'bookcase', 'chair', 'desk', 'misc',
                 'sofa', 'table', 'tool', 'wardrobe']

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        # This is the description that will appear on the datasets page.
        description=_DESCRIPTION,
        # tfds.features.FeatureConnectors
        features=tfds.features.FeaturesDict({
            'image': tfds_features.Image(shape=(None, None, 3), dtype=tf.uint8),
            'image/filename': tfds_features.Text(),
            'image/source': tfds_features.Text(),
            '2d_keypoints': tfds_features.FeaturesDict({
                'num_annotators': tf.uint8,
                'num_keypoints': tf.uint8,
                'keypoints': tfds_features.Tensor(shape=(None,),
                                                  dtype=tf.float32),
            }),
            'mask': tfds_features.Image(shape=(None, None, 1), dtype=tf.uint8),
            'model': tfg_features.TriangleMesh(),
            'model/source': tfds_features.Text(),
            '3d_keypoints': tfds_features.Tensor(shape=(None, 3),
                                                 dtype=tf.float32),
            'voxel': tfg_features.VoxelGrid(shape=(128, 128, 128)),
            'pose': tfg_features.Pose(),  # pose of object w.r.t to world.
            'camera': tfds_features.FeaturesDict({
                'parameters': tfg_features.Camera(),
                'position_with_respect_to_object': tfds_features.Tensor(
                    shape=(3,), dtype=tf.float32
                ),
                'inplane_rotation': tf.float32,
            }),
            'category': tfds_features.ClassLabel(
                num_classes=len(self.CLASS_INDEX)
            ),
            'bbox': tfds_features.BBoxFeature(),
            'truncated': tf.bool,
            'occluded': tf.bool,
            'slightly_occluded': tf.bool,
        }),
        # If there's a common (input, target) tuple from the features,
        # specify them here. They'll be used if as_supervised=True in
        # builder.as_dataset.
        supervised_keys=None,
        # Homepage of the dataset for documentation
        homepage='http://pix3d.csail.mit.edu/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""

    pix3d_dir = dl_manager.download_and_extract(
        'http://pix3d.csail.mit.edu/data/pix3d.zip')

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                'samples_directory': pix3d_dir,
                'split_file': self.TRAIN_SPLIT_IDX
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                'samples_directory': pix3d_dir,
                'split_file': self.TEST_SPLIT_IDX
            },
        ),
    ]

  def _generate_examples(self, samples_directory, split_file):
    """Yields examples.

    As Pix3D does not come with a predefined train/test split, we adopt one from
    Mesh R-CNN. The split ensures
    that the 3D models appearing in the train and test sets are disjoint.

    Args:
      samples_directory: `str`, path to the directory where Pix3D is stored.
      split_file: `str`, path to .npy file containing the indices of the current
      split.
    """

    with tf.io.gfile.GFile(os.path.join(samples_directory, 'pix3d.json'),
                           mode='r') as pix3d_index:
      pix3d = json.load(pix3d_index)

    split_samples = map(pix3d.__getitem__, np.load(split_file))

    def _build_bbox(box, img_size):
      """Create a BBox with correct order of coordinates.

      Args:
        box: Bounding box of the object as provided by Pix3d
        img_size:  size of the image, in the format of [width, height]

      Returns:
        tfds.features.BBox.
      """
      xmin, ymin, xmax, ymax = box
      width, height = img_size
      return tfds_features.BBox(ymin=ymin / height, xmin=xmin / width,
                                ymax=ymax / height, xmax=xmax / width)

    def _build_camera(f, img_size):
      """Prepare features for `Camera` FeatureConnector.

      The pose originates from the official Pix3D GitHub repository and
      describes the cameras position with respect to the scene.

      The focal length is originally provided in mm, but will be converted to
      pixel here using the fixed sensor with of 32 mm, which also originates
      from the Pix3D GitHub repository.

      Link to the official Pix3D repository:
      https://github.com/xingyuansun/pix3d.

      Args:
        f: float, denoting the focal length in mm.
        img_size: tuple of two floats, denoting the image height and width.

      Returns:
        Dictionary with all Camera Parameters.
      """
      sensor_width = 32.
      return {
          'pose': {
              'R': np.array([[-1., 0., 0.], [0., -1., 0.], [0., 0., 1.]],
                            dtype=np.float32),
              't': np.zeros(3, dtype=np.float32)
          },
          'optical_center': (img_size[0] / 2, img_size[1] / 2),
          'f': (f / sensor_width * img_size[0])
      }

    def _build_2d_keypoints(keypoints):
      """Wraps keypoint feature in dict, because TFDS does not allow more than sone unknown dimensions.

      Args:
        keypoints: Array of dimension `[N, M, 2]`, where N denotes the number of
          annotators and M is the number of 2D keypoints. Keypoints are stored
          as (origin: top left corner; +x: rightward; +y: downward);
          [-1.0, -1.0] if an annotator marked this keypoint hard to label.

      Returns:
        Dictionary containing shape and flattened keypoints.
      """
      if keypoints.ndim != 3 or keypoints.shape[-1] != 2:
        raise ValueError('2D keypoints should be in shape (N, M, 2).')

      return {
          'num_annotators': keypoints.shape[0],
          'num_keypoints': keypoints.shape[1],
          'keypoints': keypoints.ravel()
      }

    for sample in split_samples:
      example = {
          'image': os.path.join(samples_directory, sample['img']),
          'image/filename': sample['img'],
          'image/source': sample['img_source'],
          '2d_keypoints': _build_2d_keypoints(
              np.asarray(sample['2d_keypoints'],
                         dtype=np.float32)),
          'mask': os.path.join(samples_directory, sample['mask']),
          'model': os.path.join(samples_directory, sample['model']),
          'model/source': sample['model_source'],
          '3d_keypoints': np.loadtxt(
              os.path.join(samples_directory, sample['3d_keypoints']),
              dtype=np.float32),
          'voxel': {
              'path': os.path.join(samples_directory, sample['voxel']),
              'key': 'voxel'
          },
          'pose': {
              'R': sample['rot_mat'],
              't': sample['trans_mat']
          },
          'camera': {
              'parameters': _build_camera(
                  sample['focal_length'],
                  sample['img_size'],
              ),
              'position_with_respect_to_object': sample['cam_position'],
              'inplane_rotation': sample['inplane_rotation'],
          },
          'category': self.CLASS_INDEX.index(sample['category']),
          'bbox': _build_bbox(sample['bbox'], sample['img_size']),
          'truncated': sample['truncated'],
          'occluded': sample['occluded'],
          'slightly_occluded': sample['slightly_occluded']
      }

      yield sample['img'], example



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements various graph convolutions in TensorFlow."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.geometry.convolution import utils
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def feature_steered_convolution(data,
                                neighbors,
                                sizes,
                                var_u,
                                var_v,
                                var_c,
                                var_w,
                                var_b,
                                name=None):
  #  pyformat: disable
  """Implements the Feature Steered graph convolution.

  FeaStNet: Feature-Steered Graph Convolutions for 3D Shape Analysis
  Nitika Verma, Edmond Boyer, Jakob Verbeek
  CVPR 2018
  https://arxiv.org/abs/1706.05206

  The shorthands used below are
    `V`: The number of vertices.
    `C`: The number of channels in the input data.
    `D`: The number of channels in the output after convolution.
    `W`: The number of weight matrices used in the convolution.
    The input variables (`var_u`, `var_v`, `var_c`, `var_w`, `var_b`) correspond
    to the variables with the same names in the paper cited above.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    data: A `float` tensor with shape `[A1, ..., An, V, C]`.
    neighbors: A `SparseTensor` with the same type as `data` and with shape
      `[A1, ..., An, V, V]` representing vertex neighborhoods. The neighborhood
      of a vertex defines the support region for convolution. For a mesh, a
      common choice for the neighborhood of vertex i would be the vertices in
      the K-ring of i (including i itself). Each vertex must have at least one
      neighbor. For a faithful implementation of the FeaStNet convolution,
      neighbors should be a row-normalized weight matrix corresponding to the
      graph adjacency matrix with self-edges: `neighbors[A1, ..., An, i, j] > 0`
      if vertex j is a neighbor of i, and `neighbors[A1, ..., An, i, i] > 0` for
      all i, and `sum(neighbors, axis=-1)[A1, ..., An, i] == 1.0 for all i`.
      These requirements are relaxed in this implementation.
    sizes: An `int` tensor of shape `[A1, ..., An]` indicating the true input
      sizes in case of padding (`sizes=None` indicates no padding).Note that
      `sizes[A1, ..., An] <= V`. If `data` and `neighbors` are 2-D, `sizes` will
      be ignored. An example usage of `sizes`: consider an input consisting of
      three graphs G0, G1, and G2 with V0, V1, and V2 vertices respectively. The
      padded input would have the following shapes: `data.shape = [3, V, C]` and
      `neighbors.shape = [3, V, V]`, where `V = max([V0, V1, V2])`. The true
      sizes of each graph will be specified by `sizes=[V0, V1, V2]`,
      `data[i, :Vi, :]` and `neighbors[i, :Vi, :Vi]` will be the vertex and
      neighborhood data of graph Gi. The `SparseTensor` `neighbors` should have
      no nonzero entries in the padded regions.
    var_u: A 2-D tensor with shape `[C, W]`.
    var_v: A 2-D tensor with shape `[C, W]`.
    var_c: A 1-D tensor with shape `[W]`.
    var_w: A 3-D tensor with shape `[W, C, D]`.
    var_b: A 1-D tensor with shape `[D]`.
    name: A name for this op. Defaults to
      `graph_convolution_feature_steered_convolution`.

  Returns:
    Tensor with shape `[A1, ..., An, V, D]`.

  Raises:
    TypeError: if the input types are invalid.
    ValueError: if the input dimensions are invalid.
  """
  #  pyformat: enable
  with tf.compat.v1.name_scope(
      name, "graph_convolution_feature_steered_convolution",
      [data, neighbors, sizes, var_u, var_v, var_c, var_w, var_b]):
    data = tf.convert_to_tensor(value=data)
    neighbors = tf.compat.v1.convert_to_tensor_or_sparse_tensor(value=neighbors)
    if sizes is not None:
      sizes = tf.convert_to_tensor(value=sizes)
    var_u = tf.convert_to_tensor(value=var_u)
    var_v = tf.convert_to_tensor(value=var_v)
    var_c = tf.convert_to_tensor(value=var_c)
    var_w = tf.convert_to_tensor(value=var_w)
    var_b = tf.convert_to_tensor(value=var_b)

    data_ndims = data.shape.ndims
    utils.check_valid_graph_convolution_input(data, neighbors, sizes)
    shape.compare_dimensions(
        tensors=(data, var_u, var_v, var_w),
        tensor_names=("data", "var_u", "var_v", "var_w"),
        axes=(-1, 0, 0, 1))
    shape.compare_dimensions(
        tensors=(var_u, var_v, var_c, var_w),
        tensor_names=("var_u", "var_v", "var_c", "var_w"),
        axes=(1, 1, 0, 0))
    shape.compare_dimensions(
        tensors=(var_w, var_b), tensor_names=("var_w", "var_b"), axes=-1)

    # Flatten the batch dimensions and remove any vertex padding.
    if data_ndims > 2:
      if sizes is not None:
        sizes_square = tf.stack((sizes, sizes), axis=-1)
      else:
        sizes_square = None
      x_flat, unflatten = utils.flatten_batch_to_2d(data, sizes)
      adjacency = utils.convert_to_block_diag_2d(neighbors, sizes_square)
    else:
      x_flat = data
      adjacency = neighbors
    x_u = tf.matmul(x_flat, var_u)
    x_v = tf.matmul(x_flat, var_v)
    adjacency_ind_0 = adjacency.indices[:, 0]
    adjacency_ind_1 = adjacency.indices[:, 1]
    x_u_rep = tf.gather(x_u, adjacency_ind_0)
    x_v_sep = tf.gather(x_v, adjacency_ind_1)
    weights_q = tf.exp(x_u_rep + x_v_sep + tf.reshape(var_c, (1, -1)))
    weights_q_sum = tf.reduce_sum(
        input_tensor=weights_q, axis=-1, keepdims=True)
    weights_q = weights_q / weights_q_sum
    y_i_m = []
    x_sep = tf.gather(x_flat, adjacency_ind_1)
    q_m_list = tf.unstack(weights_q, axis=-1)
    w_m_list = tf.unstack(var_w, axis=0)

    x_flat_shape = tf.shape(input=x_flat)
    for q_m, w_m in zip(q_m_list, w_m_list):
      # Compute `y_i_m = sum_{j in neighborhood(i)} q_m(x_i, x_j) * w_m * x_j`.
      q_m = tf.expand_dims(q_m, axis=-1)
      p_sum = tf.math.unsorted_segment_sum(
          data=(q_m * x_sep) * tf.expand_dims(adjacency.values, -1),
          segment_ids=adjacency_ind_0,
          num_segments=x_flat_shape[0])
      y_i_m.append(tf.matmul(p_sum, w_m))
    y_out = tf.add_n(inputs=y_i_m) + tf.reshape(var_b, [1, -1])
    if data_ndims > 2:
      y_out = unflatten(y_out)
    return y_out


def edge_convolution_template(data,
                              neighbors,
                              sizes,
                              edge_function,
                              reduction,
                              edge_function_kwargs,
                              name=None):
  #  pyformat: disable
  r"""A template for edge convolutions.

  This function implements a general edge convolution for graphs of the form
  \\(y_i = \sum_{j \in \mathcal{N}(i)} w_{ij} f(x_i, x_j)\\), where
  \\(\mathcal{N}(i)\\) is the set of vertices in the neighborhood of vertex
  \\(i\\), \\(x_i \in \mathbb{R}^C\\) are the features at vertex \\(i\\),
  \\(w_{ij} \in \mathbb{R}\\) is the weight for the edge between vertex \\(i\\)
  and vertex \\(j\\), and finally
  \\(f(x_i, x_j): \mathbb{R}^{C} \times \mathbb{R}^{C} \to \mathbb{R}^{D}\\) is
  a user-supplied function.

  This template also implements the same general edge convolution described
  above with a max-reduction instead of a weighted sum.

  An example of how this template can be used is for Laplacian smoothing,
  which is defined as
  $$y_i = \frac{1}{|\mathcal{N(i)}|} \sum_{j \in \mathcal{N(i)}} x_j$$.
  `edge_convolution_template` can be used to perform Laplacian smoothing by
  setting
  $$w_{ij} = \frac{1}{|\mathcal{N(i)}|}$$, `edge_function=lambda x, y: y`,
  and `reduction='weighted'`.

  The shorthands used below are
    `V`: The number of vertices.
    `C`: The number of channels in the input data.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    data: A `float` tensor with shape `[A1, ..., An, V, C]`.
    neighbors: A `SparseTensor` with the same type as `data` and with shape
      `[A1, ..., An, V, V]` representing vertex neighborhoods. The neighborhood
      of a vertex defines the support region for convolution. The value at
      `neighbors[A1, ..., An, i, j]` corresponds to the weight \\(w_{ij}\\)
      above. Each vertex must have at least one neighbor.
    sizes: An `int` tensor of shape `[A1, ..., An]` indicating the true input
      sizes in case of padding (`sizes=None` indicates no padding). Note that
      `sizes[A1, ..., An] <= V`. If `data` and `neighbors` are 2-D, `sizes` will
      be ignored. As an example, consider an input consisting of three graphs
      G0, G1, and G2 with V0, V1, and V2 vertices respectively. The padded input
      would have the shapes `[3, V, C]`, and `[3, V, V]` for `data` and
      `neighbors` respectively, where `V = max([V0, V1, V2])`. The true sizes of
      each graph will be specified by `sizes=[V0, V1, V2]` and `data[i, :Vi, :]`
      and `neighbors[i, :Vi, :Vi]` will be the vertex and neighborhood data of
      graph Gi. The `SparseTensor` `neighbors` should have no nonzero entries in
      the padded regions.
    edge_function: A callable that takes at least two arguments of vertex
      features and returns a tensor of vertex features. `Y = f(X1, X2,
      **kwargs)`, where `X1` and `X2` have shape `[V3, C]` and `Y` must have
      shape `[V3, D], D >= 1`.
    reduction: Either 'weighted' or 'max'. Specifies the reduction over the
        neighborhood. For 'weighted', the reduction is a weighted sum as shown
        in the equation above. For 'max' the reduction is a max over features in
        which case the weights $$w_{ij}$$ are ignored.
    edge_function_kwargs: A dict containing any additional keyword arguments to
      be passed to `edge_function`.
    name: A name for this op. Defaults to
      `graph_convolution_edge_convolution_template`.

  Returns:
    Tensor with shape `[A1, ..., An, V, D]`.

  Raises:
    TypeError: if the input types are invalid.
    ValueError: if the input dimensions are invalid.
  """
  #  pyformat: enable
  with tf.compat.v1.name_scope(name,
                               "graph_convolution_edge_convolution_template",
                               [data, neighbors, sizes]):
    data = tf.convert_to_tensor(value=data)
    neighbors = tf.compat.v1.convert_to_tensor_or_sparse_tensor(value=neighbors)
    if sizes is not None:
      sizes = tf.convert_to_tensor(value=sizes)

    data_ndims = data.shape.ndims
    utils.check_valid_graph_convolution_input(data, neighbors, sizes)

    # Flatten the batch dimensions and remove any vertex padding.
    if data_ndims > 2:
      if sizes is not None:
        sizes_square = tf.stack((sizes, sizes), axis=-1)
      else:
        sizes_square = None
      x_flat, unflatten = utils.flatten_batch_to_2d(data, sizes)
      adjacency = utils.convert_to_block_diag_2d(neighbors, sizes_square)
    else:
      x_flat = data
      adjacency = neighbors

    adjacency_ind_0 = adjacency.indices[:, 0]
    adjacency_ind_1 = adjacency.indices[:, 1]
    vertex_features = tf.gather(x_flat, adjacency_ind_0)
    neighbor_features = tf.gather(x_flat, adjacency_ind_1)
    edge_features = edge_function(vertex_features, neighbor_features,
                                  **edge_function_kwargs)

    if reduction == "weighted":
      edge_features_weighted = edge_features * tf.expand_dims(
          adjacency.values, -1)
      features = tf.math.unsorted_segment_sum(
          data=edge_features_weighted,
          segment_ids=adjacency_ind_0,
          num_segments=tf.shape(input=x_flat)[0])
    elif reduction == "max":
      features = tf.math.segment_max(data=edge_features,
                                     segment_ids=adjacency_ind_0)
    else:
      raise ValueError("The reduction method must be 'weighted' or 'max'")

    features.set_shape(features.shape.merge_with(
        (tf.compat.v1.dimension_value(x_flat.shape[0]),
         tf.compat.v1.dimension_value(edge_features.shape[-1]))))

    if data_ndims > 2:
      features = unflatten(features)
    return features


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements TensorFlow As Rigid As Possible utility functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.geometry.transformation import quaternion
from tensorflow_graphics.math import vector
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def energy(vertices_rest_pose,
           vertices_deformed_pose,
           quaternions,
           edges,
           vertex_weight=None,
           edge_weight=None,
           conformal_energy=True,
           aggregate_loss=True,
           name=None):
  """Estimates an As Conformal As Possible (ACAP) fitting energy.

  For a given mesh in rest pose, this function evaluates a variant of the ACAP
  [1] fitting energy for a batch of deformed meshes. The vertex weights and edge
  weights are defined on the rest pose.

  The method implemented here is similar to [2], but with an added free variable
    capturing a scale factor per vertex.

  [1]: Yusuke Yoshiyasu, Wan-Chun Ma, Eiichi Yoshida, and Fumio Kanehiro.
  "As-Conformal-As-Possible Surface Registration." Computer Graphics Forum. Vol.
  33. No. 5. 2014.</br>
  [2]: Olga Sorkine, and Marc Alexa.
  "As-rigid-as-possible surface modeling". Symposium on Geometry Processing.
  Vol. 4. 2007.

  Note:
    In the description of the arguments, V corresponds to
      the number of vertices in the mesh, and E to the number of edges in this
      mesh.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vertices_rest_pose: A tensor of shape `[V, 3]` containing the position of
      all the vertices of the mesh in rest pose.
    vertices_deformed_pose: A tensor of shape `[A1, ..., An, V, 3]` containing
      the position of all the vertices of the mesh in deformed pose.
    quaternions: A tensor of shape `[A1, ..., An, V, 4]` defining a rigid
      transformation to apply to each vertex of the rest pose. See Section 2
      from [1] for further details.
    edges: A tensor of shape `[E, 2]` defining indices of vertices that are
      connected by an edge.
    vertex_weight: An optional tensor of shape `[V]` defining the weight
      associated with each vertex. Defaults to a tensor of ones.
    edge_weight: A tensor of shape `[E]` defining the weight of edges. Common
      choices for these weights include uniform weighting, and cotangent
      weights. Defaults to a tensor of ones.
    conformal_energy: A `bool` indicating whether each vertex is associated with
      a scale factor or not. If this parameter is True, scaling information must
      be encoded in the norm of `quaternions`. If this parameter is False, this
      function implements the energy described in [2].
    aggregate_loss: A `bool` defining whether the returned loss should be an
      aggregate measure. When True, the mean squared error is returned. When
      False, returns two losses for every edge of the mesh.
    name: A name for this op. Defaults to "as_conformal_as_possible_energy".

  Returns:
    When aggregate_loss is `True`, returns a tensor of shape `[A1, ..., An]`
    containing the ACAP energies. When aggregate_loss is `False`, returns a
    tensor of shape `[A1, ..., An, 2*E]` containing each term of the summation
    described in the equation 7 of [2].

  Raises:
    ValueError: if the shape of `vertices_rest_pose`, `vertices_deformed_pose`,
    `quaternions`, `edges`, `vertex_weight`, or `edge_weight` is not supported.
  """
  with tf.compat.v1.name_scope(name, "as_conformal_as_possible_energy", [
      vertices_rest_pose, vertices_deformed_pose, quaternions, edges,
      conformal_energy, vertex_weight, edge_weight
  ]):
    vertices_rest_pose = tf.convert_to_tensor(value=vertices_rest_pose)
    vertices_deformed_pose = tf.convert_to_tensor(value=vertices_deformed_pose)
    quaternions = tf.convert_to_tensor(value=quaternions)
    edges = tf.convert_to_tensor(value=edges)
    if vertex_weight is not None:
      vertex_weight = tf.convert_to_tensor(value=vertex_weight)
    if edge_weight is not None:
      edge_weight = tf.convert_to_tensor(value=edge_weight)

    shape.check_static(
        tensor=vertices_rest_pose,
        tensor_name="vertices_rest_pose",
        has_rank=2,
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=vertices_deformed_pose,
        tensor_name="vertices_deformed_pose",
        has_rank_greater_than=1,
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=quaternions,
        tensor_name="quaternions",
        has_rank_greater_than=1,
        has_dim_equals=(-1, 4))
    shape.compare_batch_dimensions(
        tensors=(vertices_deformed_pose, quaternions),
        last_axes=(-3, -3),
        broadcast_compatible=False)
    shape.check_static(
        tensor=edges, tensor_name="edges", has_rank=2, has_dim_equals=(-1, 2))
    tensors_with_vertices = [vertices_rest_pose,
                             vertices_deformed_pose,
                             quaternions]
    names_with_vertices = ["vertices_rest_pose",
                           "vertices_deformed_pose",
                           "quaternions"]
    axes_with_vertices = [-2, -2, -2]
    if vertex_weight is not None:
      shape.check_static(
          tensor=vertex_weight, tensor_name="vertex_weight", has_rank=1)
      tensors_with_vertices.append(vertex_weight)
      names_with_vertices.append("vertex_weight")
      axes_with_vertices.append(0)
    shape.compare_dimensions(
        tensors=tensors_with_vertices,
        axes=axes_with_vertices,
        tensor_names=names_with_vertices)
    if edge_weight is not None:
      shape.check_static(
          tensor=edge_weight, tensor_name="edge_weight", has_rank=1)
      shape.compare_dimensions(
          tensors=(edges, edge_weight),
          axes=(0, 0),
          tensor_names=("edges", "edge_weight"))

    if not conformal_energy:
      quaternions = quaternion.normalize(quaternions)
    # Extracts the indices of vertices.
    indices_i, indices_j = tf.unstack(edges, axis=-1)
    # Extracts the vertices we need per term.
    vertices_i_rest = tf.gather(vertices_rest_pose, indices_i, axis=-2)
    vertices_j_rest = tf.gather(vertices_rest_pose, indices_j, axis=-2)
    vertices_i_deformed = tf.gather(vertices_deformed_pose, indices_i, axis=-2)
    vertices_j_deformed = tf.gather(vertices_deformed_pose, indices_j, axis=-2)
    # Extracts the weights we need per term.
    weights_shape = vertices_i_rest.shape.as_list()[-2]
    if vertex_weight is not None:
      weight_i = tf.gather(vertex_weight, indices_i)
      weight_j = tf.gather(vertex_weight, indices_j)
    else:
      weight_i = weight_j = tf.ones(
          weights_shape, dtype=vertices_rest_pose.dtype)
    weight_i = tf.expand_dims(weight_i, axis=-1)
    weight_j = tf.expand_dims(weight_j, axis=-1)
    if edge_weight is not None:
      weight_ij = edge_weight
    else:
      weight_ij = tf.ones(weights_shape, dtype=vertices_rest_pose.dtype)
    weight_ij = tf.expand_dims(weight_ij, axis=-1)
    # Extracts the rotation we need per term.
    quaternion_i = tf.gather(quaternions, indices_i, axis=-2)
    quaternion_j = tf.gather(quaternions, indices_j, axis=-2)
    # Computes the energy.
    deformed_ij = vertices_i_deformed - vertices_j_deformed
    rotated_rest_ij = quaternion.rotate((vertices_i_rest - vertices_j_rest),
                                        quaternion_i)
    energy_ij = weight_i * weight_ij * (deformed_ij - rotated_rest_ij)
    deformed_ji = vertices_j_deformed - vertices_i_deformed
    rotated_rest_ji = quaternion.rotate((vertices_j_rest - vertices_i_rest),
                                        quaternion_j)
    energy_ji = weight_j * weight_ij * (deformed_ji - rotated_rest_ji)
    energy_ij_squared = vector.dot(energy_ij, energy_ij, keepdims=False)
    energy_ji_squared = vector.dot(energy_ji, energy_ji, keepdims=False)
    if aggregate_loss:
      average_energy_ij = tf.reduce_mean(
          input_tensor=energy_ij_squared, axis=-1)
      average_energy_ji = tf.reduce_mean(
          input_tensor=energy_ji_squared, axis=-1)
      return (average_energy_ij + average_energy_ji) / 2.0
    return tf.concat((energy_ij_squared, energy_ji_squared), axis=-1)

# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Type aliases for Python 3 typing."""

from typing import Union, Sequence

import numpy as np
import tensorflow as tf

Integer = Union[int, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16,
                np.uint32, np.uint64]
Float = Union[float, np.float16, np.float32, np.float64]
TensorLike = Union[Integer, Float, Sequence, np.ndarray, tf.Tensor, tf.Variable]



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements various graph pooling ops."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.geometry.convolution import utils
from tensorflow_graphics.util import export_api


def pool(data, pool_map, sizes, algorithm='max', name=None):
  #  pyformat: disable
  """Implements graph pooling.

  The features at each output vertex are computed by pooling over a subset of
  vertices in the input graph. This pooling window is specified by the input
  `pool_map`.

  The shorthands used below are
    `V1`: The number of vertices in the input data.
    `V2`: The number of vertices in the pooled output data.
    `C`: The number of channels in the data.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    data: A `float` tensor with shape `[A1, ..., An, V1, C]`.
    pool_map: A `SparseTensor` with the same type as `data` and with shape
      `[A1, ..., An, V2, V1]`. The features for an output vertex `v2` will be
      computed by pooling over the corresponding input vertices specified by
      the entries in `pool_map[A1, ..., An, v2, :]`.
    sizes: An `int` tensor of shape `[A1, ..., An, 2]` indicating the true
      input sizes in case of padding (`sizes=None` indicates no padding).
      `sizes[A1, ..., An, 0] <= V2` specifies the padding in the (pooled)
      output, and `sizes[A1, ..., An, 1] <= V1` specifies the padding in the
      input.
    algorithm: The pooling function, must be either 'max' or 'weighted'. Default
      is 'max'. For 'max' pooling, the output features are the maximum over the
      input vertices (in this case only the indices of the `SparseTensor`
      `pool_map` are used, the values are ignored). For 'weighted', the output
      features are a weighted sum of the input vertices, the weights specified
      by the values of `pool_map`.
    name: A name for this op. Defaults to 'graph_pooling_pool'.

  Returns:
    Tensor with shape `[A1, ..., An, V2, C]`.

  Raises:
    TypeError: if the input types are invalid.
    ValueError: if the input dimensions are invalid.
    ValueError: if `algorithm` is invalid.
  """
  #  pyformat: enable
  with tf.compat.v1.name_scope(
      name, 'graph_pooling_pool', [data, pool_map, sizes]):
    data = tf.convert_to_tensor(value=data)
    pool_map = tf.compat.v1.convert_to_tensor_or_sparse_tensor(value=pool_map)
    if sizes is not None:
      sizes = tf.convert_to_tensor(value=sizes)
    utils.check_valid_graph_pooling_input(data, pool_map, sizes)

    if sizes is not None:
      sizes_output, sizes_input = tf.split(sizes, 2, axis=-1)
      sizes_output = tf.squeeze(sizes_output, axis=-1)
      sizes_input = tf.squeeze(sizes_input, axis=-1)
    else:
      sizes_output = None
      sizes_input = None

    batched = data.shape.ndims > 2
    if batched:
      x_flat, _ = utils.flatten_batch_to_2d(data, sizes_input)
      pool_map_block_diagonal = utils.convert_to_block_diag_2d(pool_map, sizes)
    else:
      x_flat = data
      pool_map_block_diagonal = pool_map

    if algorithm == 'weighted':
      pooled = tf.sparse.sparse_dense_matmul(pool_map_block_diagonal, x_flat)
    elif algorithm == 'max':
      pool_groups = tf.gather(x_flat, pool_map_block_diagonal.indices[:, 1])
      pooled = tf.math.segment_max(
          data=pool_groups, segment_ids=pool_map_block_diagonal.indices[:, 0])
    else:
      raise ValueError('The pooling method must be "weighted" or "max"')

    if batched:
      if sizes_output is not None:
        pooled = utils.unflatten_2d_to_batch(pooled, sizes_output)
      else:
        output_shape = tf.concat((tf.shape(input=pool_map)[:-1], (-1,)), axis=0)
        pooled = tf.reshape(pooled, output_shape)

    return pooled


def unpool(data, pool_map, sizes, name=None):
  #  pyformat: disable
  r"""Graph upsampling by inverting the pooling map.

  Upsamples a graph by applying a pooling map in reverse. The inputs `pool_map`
  and `sizes` are the same as used for pooling:

  >>> pooled = pool(data, pool_map, sizes)
  >>> upsampled = unpool(pooled, pool_map, sizes)

  The shorthands used below are
    `V1`: The number of vertices in the input data.
    `V2`: The number of vertices in the unpooled output data.
    `C`: The number of channels in the data.

  Note:
    In the following, A1 to A3 are optional batch dimensions. Only up to three
    batch dimensions are supported due to limitations with TensorFlow's
    dense-sparse multiplication.

  Please see the documentation for `graph_pooling.pool` for a detailed
  interpretation of the inputs `pool_map` and `sizes`.

  Args:
    data: A `float` tensor with shape `[A1, ..., A3, V1, C]`.
    pool_map: A `SparseTensor` with the same type as `data` and with shape
      `[A1, ..., A3, V1, V2]`. The features for vertex `v1` are computed by
      pooling over the entries in `pool_map[A1, ..., A3, v1, :]`. This function
      applies this pooling map in reverse.
    sizes: An `int` tensor of shape `[A1, ..., A3, 2]` indicating the true
      input sizes in case of padding (`sizes=None` indicates no padding):
      `sizes[A1, ..., A3, 0] <= V1` and `sizes[A1, ..., A3, 1] <= V2`.
    name: A name for this op. Defaults to 'graph_pooling_unpool'.

  Returns:
    Tensor with shape `[A1, ..., A3, V2, C]`.

  Raises:
    TypeError: if the input types are invalid.
    ValueError: if the input dimensions are invalid.
  """
  #  pyformat: enable
  with tf.compat.v1.name_scope(
      name, 'graph_pooling_unpool', [data, pool_map, sizes]):
    data = tf.convert_to_tensor(value=data)
    pool_map = tf.compat.v1.convert_to_tensor_or_sparse_tensor(value=pool_map)
    if sizes is not None:
      sizes = tf.convert_to_tensor(value=sizes)
    utils.check_valid_graph_unpooling_input(data, pool_map, sizes)

    # Reverse pool_map and sizes.
    pool_map_ndims = pool_map.shape.ndims
    permutation = tf.concat((tf.range(pool_map_ndims - 2),
                             (pool_map_ndims - 1, pool_map_ndims - 2)),
                            axis=0)
    pool_map_transpose = tf.sparse.transpose(pool_map, permutation)
    row_sum = tf.sparse.reduce_sum(
        tf.abs(pool_map_transpose), keepdims=True, axis=-1)
    normalize_weights = tf.compat.v1.where(
        tf.equal(row_sum, 0), row_sum, 1.0 / row_sum)
    pool_map_normalize = normalize_weights * pool_map_transpose

    if sizes is not None:
      sizes = tf.reverse(sizes, axis=(-1,))

    return pool(data, pool_map_normalize, sizes)


def upsample_transposed_convolution(data,
                                    pool_map,
                                    sizes,
                                    kernel_size,
                                    transposed_convolution_op,
                                    name=None):
  #  pyformat: disable
  r"""Graph upsampling by transposed convolution.

  Upsamples a graph using a transposed convolution op. The map from input
  vertices to the upsampled graph is specified by the reverse of pool_map. The
  inputs `pool_map` and `sizes` are the same as used for pooling:

  >>> pooled = pool(data, pool_map, sizes)
  >>> upsampled = upsample_transposed_convolution(pooled, pool_map, sizes, ...)

  The shorthands used below are
    `V1`: The number of vertices in the inputs.
    `V2`: The number of vertices in the upsampled output.
    `C`: The number of channels in the inputs.

  Note:
    In the following, A1 to A3 are optional batch dimensions. Only up to three
    batch dimensions are supported due to limitations with TensorFlow's
    dense-sparse multiplication.

  Please see the documentation for `graph_pooling.pool` for a detailed
  interpretation of the inputs `pool_map` and `sizes`.

  Args:
    data: A `float` tensor with shape `[A1, ..., A3, V1, C]`.
    pool_map: A `SparseTensor` with the same type as `data` and with shape
      `[A1, ..., A3, V1, V2]`. `pool_map` will be interpreted in the same way
      as the `pool_map` argument of `graph_pooling.pool`, namely
      `v_i_map = [..., v_i, :]` are the upsampled vertices corresponding to
      vertex `v_i`. Additionally, for transposed convolution a fixed number of
      entries in each `v_i_map` (equal to `kernel_size`) are expected:
      `|v_i_map| = kernel_size`. When this is not the case, the map is either
      truncated or the last element repeated. Furthermore, upsampled vertex
      indices should not be repeated across maps otherwise the output is
      nondeterministic. Specifically, to avoid nondeterminism we must have
      `intersect([a1, ..., an, v_i, :],[a1, ..., a3, v_j, :]) = {}, i != j`.
    sizes: An `int` tensor of shape `[A1, ..., A3, 2]` indicating the true
      input sizes in case of padding (`sizes=None` indicates no padding):
      `sizes[A1, ..., A3, 0] <= V1` and `sizes[A1, ..., A3, 1] <= V2`.
    kernel_size: The kernel size for transposed convolution.
    transposed_convolution_op: A callable transposed convolution op with the
      form `y = transposed_convolution_op(x)`, where `x` has shape
      `[1, 1, D1, C]` and `y` must have shape `[1, 1, kernel_size * D1, C]`.
      `transposed_convolution_op` maps each row of `x` to `kernel_size` rows
      in `y`. An example:
      `transposed_convolution_op = tf.keras.layers.Conv2DTranspose(
          filters=C, kernel_size=(1, kernel_size), strides=(1, kernel_size),
          padding='valid', ...)
    name: A name for this op. Defaults to
      'graph_pooling_upsample_transposed_convolution'.

  Returns:
    Tensor with shape `[A1, ..., A3, V2, C]`.

  Raises:
    TypeError: if the input types are invalid.
    TypeError: if `transposed_convolution_op` is not a callable.
    ValueError: if the input dimensions are invalid.
  """
  #  pyformat: enable
  with tf.compat.v1.name_scope(
      name, 'graph_pooling_upsample_transposed_convolution',
      [data, pool_map, sizes]):
    data = tf.convert_to_tensor(value=data)
    pool_map = tf.compat.v1.convert_to_tensor_or_sparse_tensor(value=pool_map)
    if sizes is not None:
      sizes = tf.convert_to_tensor(value=sizes)

    utils.check_valid_graph_unpooling_input(data, pool_map, sizes)
    if not callable(transposed_convolution_op):
      raise TypeError("'transposed_convolution_op' must be callable.")

    if sizes is not None:
      sizes_input, sizes_output = tf.split(sizes, 2, axis=-1)
      sizes_input = tf.squeeze(sizes_input, axis=-1)
      sizes_output = tf.squeeze(sizes_output, axis=-1)
    else:
      sizes_input = None
      sizes_output = None

    num_features = tf.compat.v1.dimension_value(data.shape[-1])
    batched = data.shape.ndims > 2
    if batched:
      x_flat, _ = utils.flatten_batch_to_2d(data, sizes_input)
      pool_map_block_diagonal = utils.convert_to_block_diag_2d(pool_map, sizes)
    else:
      x_flat = data
      pool_map_block_diagonal = pool_map

    x_flat = tf.expand_dims(tf.expand_dims(x_flat, 0), 0)
    x_upsample = transposed_convolution_op(x_flat)

    # Map each upsampled vertex into its correct position based on pool_map.
    # Select 'kernel_size' neighbors for each input vertex. Truncate or repeat
    # as necessary.
    ragged = tf.RaggedTensor.from_value_rowids(
        pool_map_block_diagonal.indices[:, 1],
        pool_map_block_diagonal.indices[:, 0])
    # Take up to the first 'kernel_size' entries.
    ragged_k = ragged[:, :kernel_size]
    # Fill rows with less than 'kernel_size' entries by repeating the last
    # entry.
    last = ragged_k[:, -1:].flat_values
    num_repeat = kernel_size - ragged_k.row_lengths()
    sum_num_repeat = tf.reduce_sum(input_tensor=num_repeat)
    ones_ragged = tf.RaggedTensor.from_row_lengths(
        tf.ones((sum_num_repeat,), dtype=last.dtype), num_repeat)
    repeat = ones_ragged * tf.expand_dims(last, -1)
    padded = tf.concat([ragged_k, repeat], axis=1)
    pool_map_dense = tf.reshape(padded.flat_values, (-1, kernel_size))

    # Map rows of 'x_upsample' to positions indicated by the
    # indices 'pool_map_dense'.
    up_scatter_indices = tf.expand_dims(tf.reshape(pool_map_dense, (-1,)), -1)
    up_row = tf.reshape(tf.cast(up_scatter_indices, tf.int64), (-1,))
    up_column = tf.range(tf.shape(input=up_row, out_type=tf.dtypes.int64)[0])
    scatter_indices = tf.concat(
        (tf.expand_dims(up_row, -1),
         tf.expand_dims(up_column, -1)), axis=1)

    scatter_values = tf.ones_like(up_row, dtype=x_upsample.dtype)
    scatter_shape = tf.reduce_max(input_tensor=scatter_indices, axis=0) + 1

    scatter = tf.SparseTensor(scatter_indices,
                              scatter_values,
                              scatter_shape)
    scatter = tf.sparse.reorder(scatter)
    row_sum = tf.sparse.reduce_sum(tf.abs(scatter), keepdims=True, axis=-1)
    row_sum = tf.compat.v1.where(tf.equal(row_sum, 0.), row_sum, 1.0 / row_sum)
    scatter = row_sum * scatter
    x_upsample = tf.sparse.sparse_dense_matmul(scatter, x_upsample[0, 0, :, :])

    if batched:
      if sizes_output is not None:
        x_upsample = utils.unflatten_2d_to_batch(x_upsample, sizes_output)
      else:
        output_shape = tf.concat((tf.shape(input=pool_map)[:-2],
                                  tf.shape(input=pool_map)[-1:],
                                  (num_features,)), axis=0)
        x_upsample = tf.reshape(x_upsample, output_shape)

    return x_upsample

# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the intersection-over-union metric."""

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.nn.metric import intersection_over_union
from tensorflow_graphics.util import test_case


def random_tensor(tensor_shape):
  return np.random.uniform(low=0.0, high=1.0, size=tensor_shape)


def random_tensor_shape():
  tensor_size = np.random.randint(5) + 1
  return np.random.randint(1, 10, size=(tensor_size)).tolist()


class IntersectionOverUnionTest(test_case.TestCase):

  @parameterized.parameters(
      # 1 dimensional grid.
      ([1., 0, 0, 1, 1, 0, 1], \
       [1., 0, 1, 1, 1, 1, 0], 3. / 6.),
      # 2 dimensional grid.
      ([[1., 0, 1], [0, 0, 1], [0, 1, 1]], \
       [[0., 1, 1], [1, 1, 1], [0, 0, 1]], 3. / 8.),
      ([[0., 0, 1], [0, 0, 0]], \
       [[1., 1, 0], [0, 0, 1]], 0.),
      # Returns 1 if the prediction and ground-truth are all zeros.
      ([[0., 0, 0], [0, 0, 0]], \
       [[0., 0, 0], [0, 0, 0]], 1.),
  )
  def test_evaluate_preset(self, ground_truth, predictions, expected_iou):
    tensor_shape = random_tensor_shape()

    grid_size = np.array(ground_truth).ndim
    ground_truth_labels = np.tile(ground_truth, tensor_shape + [1] * grid_size)
    predicted_labels = np.tile(predictions, tensor_shape + [1] * grid_size)
    expected = np.tile(expected_iou, tensor_shape)

    result = intersection_over_union.evaluate(ground_truth_labels,
                                              predicted_labels, grid_size)

    self.assertAllClose(expected, result)

  @parameterized.parameters(
      # Exception is raised since not all the value are in the range [0, 1]
      (r"0.* 0\.7.* 0", [0., 0.7, 0], [0., 1.0, 0.0], 1),
      (r"0\.3.* 1.* 0.*", [0., 1.0, 0], [0.3, 1.0, 0.0], 1),
      # Grid size must be less or equal to min_inputs_dimensions.
      (r"Invalid reduction dimension .*\-2 for input with 1 dimension",
       [1., 0.], [[0., 1.0], [1., 1.0]], 2),
  )
  def test_evaluate_invalid_argument_exception_raised(self, error_msg,
                                                      ground_truth, predictions,
                                                      grid_size):
    with self.assertRaisesRegexp((tf.errors.InvalidArgumentError, ValueError),
                                 error_msg):
      self.evaluate(
          intersection_over_union.evaluate(ground_truth, predictions,
                                           grid_size))

  @parameterized.parameters(
      ("Not all batch dimensions are broadcast-compatible.", (1, 5, 3), (4, 3)),
      ("Not all batch dimensions are broadcast-compatible.", (3, 4), (2, 4, 5)),
  )
  def test_evaluate_shape_exception_raised(self, error_msg, *shape):
    """Tests that the shape exception is raised."""
    self.assert_exception_is_raised(intersection_over_union.evaluate, error_msg,
                                    shape)

  @parameterized.parameters(
      ((1, 5, 3), (2, 5, 1)),
      ((None, 2, 6), (4, 2, None)),
      ((3, 5, 8, 2), (3, 1, 1, 2)),
  )
  def test_evaluate_shape_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(intersection_over_union.evaluate,
                                        shapes)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""This module implements image transformation functionalities."""

import enum
from typing import Optional

import tensorflow as tf
from tensorflow_addons import image as tfa_image

from tensorflow_graphics.geometry.representation import grid
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape
from tensorflow_graphics.util import type_alias


class ResamplingType(enum.Enum):
  NEAREST = 0
  BILINEAR = 1


class BorderType(enum.Enum):
  ZERO = 0
  DUPLICATE = 1


class PixelType(enum.Enum):
  INTEGER = 0
  HALF_INTEGER = 1


def sample(image: type_alias.TensorLike,
           warp: type_alias.TensorLike,
           resampling_type: ResamplingType = ResamplingType.BILINEAR,
           border_type: BorderType = BorderType.ZERO,
           pixel_type: PixelType = PixelType.HALF_INTEGER,
           name: Optional[str] = None) -> tf.Tensor:
  """Samples an image at user defined coordinates.

  Note:
    The warp maps target to source. In the following, A1 to An are optional
    batch dimensions.

  Args:
    image: A tensor of shape `[B, H_i, W_i, C]`, where `B` is the batch size,
      `H_i` the height of the image, `W_i` the width of the image, and `C` the
      number of channels of the image.
    warp: A tensor of shape `[B, A_1, ..., A_n, 2]` containing the x and y
      coordinates at which sampling will be performed. The last dimension must
      be 2, representing the (x, y) coordinate where x is the index for width
      and y is the index for height.
   resampling_type: Resampling mode. Supported values are
     `ResamplingType.NEAREST` and `ResamplingType.BILINEAR`.
    border_type: Border mode. Supported values are `BorderType.ZERO` and
      `BorderType.DUPLICATE`.
    pixel_type: Pixel mode. Supported values are `PixelType.INTEGER` and
      `PixelType.HALF_INTEGER`.
    name: A name for this op. Defaults to "sample".

  Returns:
    Tensor of sampled values from `image`. The output tensor shape
    is `[B, A_1, ..., A_n, C]`.

  Raises:
    ValueError: If `image` has rank != 4. If `warp` has rank < 2 or its last
    dimension is not 2. If `image` and `warp` batch dimension does not match.
  """
  with tf.name_scope(name or "sample"):
    image = tf.convert_to_tensor(image, name="image")
    warp = tf.convert_to_tensor(warp, name="warp")

    shape.check_static(image, tensor_name="image", has_rank=4)
    shape.check_static(
        warp,
        tensor_name="warp",
        has_rank_greater_than=1,
        has_dim_equals=(-1, 2))
    shape.compare_batch_dimensions(
        tensors=(image, warp), last_axes=0, broadcast_compatible=False)

    if pixel_type == PixelType.HALF_INTEGER:
      warp -= 0.5

    if resampling_type == ResamplingType.NEAREST:
      warp = tf.math.round(warp)

    if border_type == BorderType.DUPLICATE:
      image_size = tf.cast(tf.shape(image)[1:3], dtype=warp.dtype)
      height, width = tf.unstack(image_size, axis=-1)
      warp_x, warp_y = tf.unstack(warp, axis=-1)
      warp_x = tf.clip_by_value(warp_x, 0.0, width - 1.0)
      warp_y = tf.clip_by_value(warp_y, 0.0, height - 1.0)
      warp = tf.stack((warp_x, warp_y), axis=-1)

    return tfa_image.resampler(image, warp)


def perspective_transform(
    image: type_alias.TensorLike,
    transform_matrix: type_alias.TensorLike,
    output_shape: Optional[type_alias.TensorLike] = None,
    resampling_type: ResamplingType = ResamplingType.BILINEAR,
    border_type: BorderType = BorderType.ZERO,
    pixel_type: PixelType = PixelType.HALF_INTEGER,
    name: Optional[str] = None,
) -> tf.Tensor:
  """Applies a projective transformation to an image.

  The projective transformation is represented by a 3 x 3 matrix
  [[a0, a1, a2], [b0, b1, b2], [c0, c1, c2]], mapping a point `[x, y]` to a
  transformed point
  `[x', y'] = [(a0 x + a1 y + a2) / k, (b0 x + b1 y + b2) / k]`, where
  `k = c0 x + c1 y + c2`.

  Note:
      The transformation matrix maps target to source by transforming output
      points to input points.

  Args:
    image: A tensor of shape `[B, H_i, W_i, C]`, where `B` is the batch size,
      `H_i` the height of the image, `W_i` the width of the image, and `C` the
      number of channels of the image.
    transform_matrix: A tensor of shape `[B, 3, 3]` containing projective
      transform matrices. The transformation maps target to source by
      transforming output points to input points.
    output_shape: The heigh `H_o` and width `W_o` output dimensions after the
      transform. If None, output is the same size as input image.
    resampling_type: Resampling mode. Supported values are
      `ResamplingType.NEAREST` and `ResamplingType.BILINEAR`.
    border_type: Border mode. Supported values are `BorderType.ZERO` and
      `BorderType.DUPLICATE`.
    pixel_type: Pixel mode. Supported values are `PixelType.INTEGER` and
      `PixelType.HALF_INTEGER`.
    name: A name for this op. Defaults to "perspective_transform".

  Returns:
    A tensor of shape `[B, H_o, W_o, C]` containing transformed images.

  Raises:
    ValueError: If `image` has rank != 4. If `transform_matrix` has rank < 3 or
    its last two dimensions are not 3. If `image` and `transform_matrix` batch
    dimension does not match.
  """
  with tf.name_scope(name or "perspective_transform"):
    image = tf.convert_to_tensor(image, name="image")
    transform_matrix = tf.convert_to_tensor(
        transform_matrix, name="transform_matrix")
    output_shape = tf.shape(
        image)[-3:-1] if output_shape is None else tf.convert_to_tensor(
            output_shape, name="output_shape")

    shape.check_static(image, tensor_name="image", has_rank=4)
    shape.check_static(
        transform_matrix,
        tensor_name="transform_matrix",
        has_rank=3,
        has_dim_equals=((-1, 3), (-2, 3)))
    shape.compare_batch_dimensions(
        tensors=(image, transform_matrix),
        last_axes=0,
        broadcast_compatible=False)

    dtype = image.dtype
    zero = tf.cast(0.0, dtype)
    height, width = tf.unstack(output_shape, axis=-1)
    warp = grid.generate(
        starts=(zero, zero),
        stops=(tf.cast(width, dtype) - 1.0, tf.cast(height, dtype) - 1.0),
        nums=(width, height))
    warp = tf.transpose(warp, perm=[1, 0, 2])

    if pixel_type == PixelType.HALF_INTEGER:
      warp += 0.5

    padding = [[0, 0] for _ in range(warp.shape.ndims)]
    padding[-1][-1] = 1
    warp = tf.pad(
        tensor=warp, paddings=padding, mode="CONSTANT", constant_values=1.0)

    warp = warp[..., tf.newaxis]
    transform_matrix = transform_matrix[:, tf.newaxis, tf.newaxis, ...]
    warp = tf.linalg.matmul(transform_matrix, warp)
    warp = warp[..., 0:2, 0] / warp[..., 2, :]

    return sample(image, warp, resampling_type, border_type, pixel_type)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow.graphics slerp interpolation module.

  Spherical linear interpolation (slerp) is defined for both quaternions and for
  regular M-D vectors, and act slightly differently because of inherent
  ambiguity of quaternions. This module has two functions returning the
  interpolation weights for quaternions (quaternion_weights) and for vectors
  (vector_weights), which can then be used in a weighted sum to calculate the
  final interpolated quaternions and vectors. A helper interpolate function is
  also provided.

  The main differences between two methods are:
  vector_weights:
    can get any M-D tensor as input,
    does not expect normalized vectors as input,
    returns unnormalized outputs (in general) for unnormalized inputs.

  quaternion_weights:
    expects M-D tensors with a last dimension of 4,
    assumes normalized input,
    checks for ambiguity by looking at the angle between quaternions,
    returns normalized quaternions naturally.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import enum
import tensorflow as tf

from tensorflow_graphics.math import vector
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import shape


class InterpolationType(enum.Enum):
  """Defines interpolation methods for slerp module."""
  VECTOR = 0
  QUATERNION = 1


def _safe_dot(vector1, vector2, eps):
  """Calculates dot product while ensuring it is in the range [-1, 1]."""
  dot_product = vector.dot(vector1, vector2)
  # Safely shrink to make sure machine precision does not cause the dot
  # product to be outside the [-1.0, 1.0] range.
  return safe_ops.safe_shrink(
      vector=dot_product, minval=-1.0, maxval=1.0, open_bounds=False, eps=eps)


def interpolate(vector1,
                vector2,
                percent,
                method=InterpolationType.QUATERNION,
                eps=None,
                name=None):
  """Applies slerp to vectors or quaternions.

  Args:
    vector1: A tensor of shape `[A1, ... , An, M]`, which stores a normalized
      vector in its last dimension.
    vector2: A tensor of shape `[A1, ... , An, M]`, which stores a normalized
      vector in its last dimension.
    percent: A `float` or a tensor with shape broadcastable to the shape of
      input vectors.
    method: An enumerated constant from the class InterpolationType, which is
      either InterpolationType.QUATERNION (default) if the input vectors are 4-D
      quaternions, or InterpolationType.VECTOR if they are regular M-D vectors.
    eps: A small float for operation safety. If left None, its value is
      automatically selected using dtype of input vectors.
    name: A name for this op. Defaults to "vector_weights" or
      "quaternion_weights" depending on the method.

  Returns:
    A tensor of shape [A1, ... , An, M]` which stores the result of the
    interpolation.

  Raises:
    ValueError: if method is not amongst enumerated constants defined in
      InterpolationType.
  """
  if method == InterpolationType.QUATERNION:
    weight1, weight2 = quaternion_weights(
        vector1, vector2, percent, eps=eps, name=name)
  elif method == InterpolationType.VECTOR:
    weight1, weight2 = vector_weights(
        vector1, vector2, percent, eps=eps, name=name)
  else:
    raise ValueError("Unknown interpolation type supplied.")
  return interpolate_with_weights(vector1, vector2, weight1, weight2)


def interpolate_with_weights(vector1, vector2, weight1, weight2, name=None):
  """Interpolates vectors by taking their weighted sum.

  Interpolation for all variants of slerp is a simple weighted sum over inputs.
  Therefore this function simply returns weight1 * vector1 + weight2 * vector2.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vector1: A tensor of shape `[A1, ... , An, M]`, which stores a normalized
      vector in its last dimension.
    vector2: A tensor of shape `[A1, ... , An, M]`, which stores a normalized
      vector in its last dimension.
    weight1: A `float` or a tensor describing weights for the `vector1` and with
      a shape broadcastable to the shape of the input vectors.
    weight2: A `float` or a tensor describing weights for the `vector2` and with
      a shape broadcastable to the shape of the input vectors.
    name: A name for this op. Defaults to "interpolate_with_weights".

  Returns:
    A tensor of shape `[A1, ... , An, M]` containing the result of the
    interpolation.
  """
  with tf.compat.v1.name_scope(name, "interpolate_with_weights",
                               [vector1, vector2, weight1, weight2]):
    return weight1 * vector1 + weight2 * vector2


def quaternion_weights(quaternion1, quaternion2, percent, eps=None, name=None):
  """Calculates slerp weights for two normalized quaternions.

  Given a percent and two normalized quaternions, this function returns the
  slerp weights. It can also produce extrapolation weights when percent is
  outside of the [0, 1] range. It reduces to lerp when input quaternions are
  almost parallel or anti-parallel. Input quaternions are assumed to be
  normalized. The tf.graphics debug flag TFG_ADD_ASSERTS_TO_GRAPH defined
  in tfg_flags.py can be set to add assertions to the graph that check whether
  the inputs are normalized, and whether Inf or Nan values are produced.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    quaternion1: A tensor of shape `[A1, ... , An, 4]` storing normalized
      quaternions in its last dimension.
    quaternion2: A tensor of shape `[A1, ... , An, 4]` storing normalized
      quaternions in its last dimension.
    percent: A `float` or a tensor with a shape broadcastable to the shape `[A1,
      ... , An]`.
    eps: A `float` used to make operations safe. When left as None, the function
      automatically picks the best epsilon based on the dtype and the operation.
    name: A name for this op. Defaults to "quaternion_weights".

  Raises:
    ValueError: If the shapes of quaternions do not match, if the last
      dimensions of quaternions are not 4, or if percent is neither a float, nor
      a tensor with last dimension 1.

  Returns:
    Two tensors of shape `[A1, ... , An, 1]` each, which are the two slerp
      weights for each quaternion.
  """
  with tf.compat.v1.name_scope(name, "quaternion_weights",
                               [quaternion1, quaternion2, percent]):
    quaternion1 = tf.convert_to_tensor(value=quaternion1)
    quaternion2 = tf.convert_to_tensor(value=quaternion2)
    percent = tf.convert_to_tensor(value=percent, dtype=quaternion1.dtype)

    if percent.shape.ndims == 0:
      percent = tf.expand_dims(percent, axis=0)
    shape.check_static(
        tensor=quaternion1, tensor_name="quaternion1", has_dim_equals=(-1, 4))
    shape.check_static(
        tensor=quaternion2, tensor_name="quaternion2", has_dim_equals=(-1, 4))
    shape.compare_batch_dimensions(
        tensors=(quaternion1, quaternion2, percent),
        last_axes=(-2, -2, -1),
        broadcast_compatible=True,
        tensor_names=("quaternion1", "quaternion2", "percent"))
    quaternion1 = asserts.assert_normalized(quaternion1)
    quaternion2 = asserts.assert_normalized(quaternion2)

    dot_product = _safe_dot(quaternion1, quaternion2, eps)

    # Take the shorter path
    theta = tf.acos(tf.abs(dot_product))

    # safe_sinpx_div_sinx returns p for very small x, which means slerp reduces
    # to lerp automatically.
    scale1 = safe_ops.safe_sinpx_div_sinx(theta, 1.0 - percent, eps)
    scale2 = safe_ops.safe_sinpx_div_sinx(theta, percent, eps)

    # Flip the sign of scale1 if quaternions are in different hemispheres.
    # tf.sign can make scale1 zero if quaternions are orthogonal.
    scale1 *= safe_ops.nonzero_sign(dot_product)
    return scale1, scale2


def vector_weights(vector1, vector2, percent, eps=None, name=None):
  """Spherical linear interpolation (slerp) between two unnormalized vectors.

  This function applies geometric slerp to unnormalized vectors by first
  normalizing them to return the interpolation weights. It reduces to lerp when
  input vectors are exactly anti-parallel.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vector1: A tensor of shape `[A1, ... , An, M]`, which stores a normalized
      vector in its last dimension.
    vector2: A tensor of shape `[A1, ... , An, M]`, which stores a normalized
      vector in its last dimension.
    percent: A `float` or tensor with shape broadcastable to the shape of input
      vectors.
    eps: A small float for operation safety. If left None, its value is
      automatically selected using dtype of input vectors.
    name: A name for this op. Defaults to "vector_weights".

  Raises:
    ValueError: if the shape of `vector1`, `vector2`, or `percent` is not
      supported.

  Returns:
    Two tensors of shape `[A1, ... , An, 1]`, representing interpolation weights
    for each input vector.
  """
  with tf.compat.v1.name_scope(name, "vector_weights",
                               [vector1, vector2, percent]):
    vector1 = tf.convert_to_tensor(value=vector1)
    vector2 = tf.convert_to_tensor(value=vector2)
    percent = tf.convert_to_tensor(value=percent, dtype=vector1.dtype)

    if percent.shape.ndims == 0:
      percent = tf.expand_dims(percent, axis=0)
    shape.compare_dimensions(
        tensors=(vector1, vector2),
        axes=-1,
        tensor_names=("vector1", "vector2"))
    shape.compare_batch_dimensions(
        tensors=(vector1, vector2, percent),
        last_axes=(-2, -2, -1),
        broadcast_compatible=True,
        tensor_names=("vector1", "vector2", "percent"))
    normalized1 = tf.nn.l2_normalize(vector1, axis=-1)
    normalized2 = tf.nn.l2_normalize(vector2, axis=-1)

    dot_product = _safe_dot(normalized1, normalized2, eps)

    theta = tf.acos(dot_product)
    scale1 = safe_ops.safe_sinpx_div_sinx(theta, 1.0 - percent, eps)
    scale2 = safe_ops.safe_sinpx_div_sinx(theta, percent, eps)
    return scale1, scale2


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow ray utility functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.math import vector
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def triangulate(startpoints, endpoints, weights, name=None):
  """Triangulates 3d points by miminizing the sum of squared distances to rays.

  The rays are defined by their start points and endpoints. At least two rays
  are required to triangulate any given point. Contrary to the standard
  reprojection-error metric, the sum of squared distances to rays can be
  minimized in a closed form.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    startpoints: A tensor of ray start points with shape `[A1, ..., An, V, 3]`,
      the number of rays V around which the solution points live should be
      greater or equal to 2, otherwise triangulation is impossible.
    endpoints: A tensor of ray endpoints with shape `[A1, ..., An, V, 3]`, the
      number of rays V around which the solution points live should be greater
      or equal to 2, otherwise triangulation is impossible. The `endpoints`
      tensor should have the same shape as the `startpoints` tensor.
    weights: A tensor of ray weights (certainties) with shape `[A1, ..., An,
      V]`. Weights should have all positive entries. Weight should have at least
      two non-zero entries for each point (at least two rays should have
      certainties > 0).
    name: A name for this op. The default value of None means "ray_triangulate".

  Returns:
    A tensor of triangulated points with shape `[A1, ..., An, 3]`.

  Raises:
    ValueError: If the shape of the arguments is not supported.
  """
  with tf.compat.v1.name_scope(name, "ray_triangulate",
                               [startpoints, endpoints, weights]):
    startpoints = tf.convert_to_tensor(value=startpoints)
    endpoints = tf.convert_to_tensor(value=endpoints)
    weights = tf.convert_to_tensor(value=weights)

    shape.check_static(
        tensor=startpoints,
        tensor_name="startpoints",
        has_rank_greater_than=1,
        has_dim_equals=(-1, 3),
        has_dim_greater_than=(-2, 1))
    shape.check_static(
        tensor=endpoints,
        tensor_name="endpoints",
        has_rank_greater_than=1,
        has_dim_equals=(-1, 3),
        has_dim_greater_than=(-2, 1))
    shape.compare_batch_dimensions(
        tensors=(startpoints, endpoints, weights),
        last_axes=(-2, -2, -1),
        broadcast_compatible=False)
    weights = asserts.assert_all_above(weights, 0.0, open_bound=False)
    weights = asserts.assert_at_least_k_non_zero_entries(weights, k=2)

    left_hand_side_list = []
    right_hand_side_list = []
    # TODO(b/130892100): Replace the inefficient for loop and add comments here.
    for ray_id in range(weights.shape[-1]):
      weights_single_ray = weights[..., ray_id]
      startpoints_single_ray = startpoints[..., ray_id, :]
      endpoints_singleview = endpoints[..., ray_id, :]
      ray = endpoints_singleview - startpoints_single_ray
      ray = tf.nn.l2_normalize(ray, axis=-1)
      ray_x, ray_y, ray_z = tf.unstack(ray, axis=-1)
      zeros = tf.zeros_like(ray_x)
      cross_product_matrix = tf.stack(
          (zeros, -ray_z, ray_y, ray_z, zeros, -ray_x, -ray_y, ray_x, zeros),
          axis=-1)
      cross_product_matrix_shape = tf.concat(
          (tf.shape(input=cross_product_matrix)[:-1], (3, 3)), axis=-1)
      cross_product_matrix = tf.reshape(
          cross_product_matrix, shape=cross_product_matrix_shape)
      weights_single_ray = tf.expand_dims(weights_single_ray, axis=-1)
      weights_single_ray = tf.expand_dims(weights_single_ray, axis=-1)
      left_hand_side = weights_single_ray * cross_product_matrix
      left_hand_side_list.append(left_hand_side)
      dot_product = tf.matmul(cross_product_matrix,
                              tf.expand_dims(startpoints_single_ray, axis=-1))
      right_hand_side = weights_single_ray * dot_product
      right_hand_side_list.append(right_hand_side)
    left_hand_side_multi_rays = tf.concat(left_hand_side_list, axis=-2)
    right_hand_side_multi_rays = tf.concat(right_hand_side_list, axis=-2)
    points = tf.linalg.lstsq(left_hand_side_multi_rays,
                             right_hand_side_multi_rays)
    points = tf.squeeze(points, axis=-1)

    return points


# TODO(b/130893491): Add batch support for radii and return [A1, ... , 3, 2].
def intersection_ray_sphere(sphere_center,
                            sphere_radius,
                            ray,
                            point_on_ray,
                            name=None):
  """Finds positions and surface normals where the sphere and the ray intersect.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    sphere_center: A tensor of shape `[3]` representing the 3d sphere center.
    sphere_radius: A tensor of shape `[1]` containing a strictly positive value
      defining the radius of the sphere.
    ray: A tensor of shape `[A1, ..., An, 3]` containing normalized 3D vectors.
    point_on_ray: A tensor of shape `[A1, ..., An, 3]`.
    name: A name for this op. The default value of None means
      "ray_intersection_ray_sphere".

  Returns:
    A tensor of shape `[2, A1, ..., An, 3]` containing the position of the
    intersections, and a tensor of shape `[2, A1, ..., An, 3]` the associated
    surface normals at that point. Both tensors contain NaNs when there is no
    intersections. The first dimension of the returned tensor provides access to
    the first and second intersections of the ray with the sphere.

  Raises:
    ValueError: if the shape of `sphere_center`, `sphere_radius`, `ray` or
      `point_on_ray` is not supported.
    tf.errors.InvalidArgumentError: If `ray` is not normalized.
  """
  with tf.compat.v1.name_scope(
      name, "ray_intersection_ray_sphere",
      [sphere_center, sphere_radius, ray, point_on_ray]):
    sphere_center = tf.convert_to_tensor(value=sphere_center)
    sphere_radius = tf.convert_to_tensor(value=sphere_radius)
    ray = tf.convert_to_tensor(value=ray)
    point_on_ray = tf.convert_to_tensor(value=point_on_ray)

    shape.check_static(
        tensor=sphere_center,
        tensor_name="sphere_center",
        has_rank=1,
        has_dim_equals=(0, 3))
    shape.check_static(
        tensor=sphere_radius,
        tensor_name="sphere_radius",
        has_rank=1,
        has_dim_equals=(0, 1))
    shape.check_static(tensor=ray, tensor_name="ray", has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=point_on_ray, tensor_name="point_on_ray", has_dim_equals=(-1, 3))
    shape.compare_batch_dimensions(
        tensors=(ray, point_on_ray),
        last_axes=(-2, -2),
        broadcast_compatible=False)
    sphere_radius = asserts.assert_all_above(
        sphere_radius, 0.0, open_bound=True)
    ray = asserts.assert_normalized(ray)

    vector_sphere_center_to_point_on_ray = sphere_center - point_on_ray
    distance_sphere_center_to_point_on_ray = tf.norm(
        tensor=vector_sphere_center_to_point_on_ray, axis=-1, keepdims=True)
    distance_projection_sphere_center_on_ray = vector.dot(
        vector_sphere_center_to_point_on_ray, ray)
    closest_distance_sphere_center_to_ray = tf.sqrt(
        tf.square(distance_sphere_center_to_point_on_ray) -
        tf.pow(distance_projection_sphere_center_on_ray, 2))
    half_secant_length = tf.sqrt(
        tf.square(sphere_radius) -
        tf.square(closest_distance_sphere_center_to_ray))
    distances = tf.stack(
        (distance_projection_sphere_center_on_ray - half_secant_length,
         distance_projection_sphere_center_on_ray + half_secant_length),
        axis=0)
    intersections_points = distances * ray + point_on_ray
    normals = tf.math.l2_normalize(
        intersections_points - sphere_center, axis=-1)
    return intersections_points, normals


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for reading and writing EXR image files as numpy arrays."""

# pylint: disable=c-extension-no-member

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import Imath
import numpy as np
import OpenEXR


_np_to_exr = {
    np.float16: Imath.PixelType.HALF,
    np.float32: Imath.PixelType.FLOAT,
    np.uint32: Imath.PixelType.UINT,
}
_exr_to_np = dict(zip(_np_to_exr.values(), _np_to_exr.keys()))


def channels_to_ndarray(exr, channel_names):
  """Copies channels from an OpenEXR.InputFile into a numpy array.

  If the EXR image is of size (width, height), the result will be a numpy array
  of shape (height, width, len(channel_names)), where the last dimension holds
  the channels in the order they were specified in channel_names. The requested
  channels must all have the same datatype.

  Args:
    exr: An OpenEXR.InputFile that is already open.
    channel_names: A list of strings naming the channels to read.

  Returns:
    A numpy ndarray.

  Raises:
    ValueError: If the channels have different datatypes.
    RuntimeError: If a channel has an unknown type.
  """
  channels_header = exr.header()['channels']
  window = exr.header()['dataWindow']
  width = window.max.x - window.min.x + 1
  height = window.max.y - window.min.y + 1

  def read_channel(channel):
    """Reads a single channel from the EXR."""
    channel_type = channels_header[channel].type
    try:
      numpy_type = _exr_to_np[channel_type.v]
    except KeyError:
      raise RuntimeError('Unknown EXR channel type: %s' % str(channel_type))
    flat_buffer = np.frombuffer(exr.channel(channel), numpy_type)
    return np.reshape(flat_buffer, [height, width])

  channels = [read_channel(c) for c in channel_names]
  if any([channels[0].dtype != c.dtype for c in channels[1:]]):
    raise ValueError('Channels have mixed datatypes: %s' %
                     ', '.join([str(c.dtype) for c in channels]))
  # Stack the arrays so that the channels dimension is the last (fastest
  # changing) dimension.
  return np.stack(channels, axis=-1)


def read_exr(filename, channel_names=None):
  """Opens an EXR file and copies the requested channels into an ndarray.

  The Python OpenEXR wrapper uses a dictionary for the channel header, so the
  ordering of the channels in the underlying file is lost. If channel_names is
  not passed, this function orders the output channels with any present RGBA
  channels first, followed by the remaining channels in alphabetical order.
  By convention, RGBA channels are named 'R', 'G', 'B', 'A', so this function
  looks for those strings.

  Args:
    filename: The name of the EXR file.
    channel_names: A list of strings naming the channels to read. If None, all
      channels will be read.

  Returns:
    A numpy array containing the image data, and a list of the corresponding
      channel names.
  """
  exr = OpenEXR.InputFile(filename)
  if channel_names is None:
    remaining_channel_names = list(exr.header()['channels'].keys())
    conventional_rgba_names = ['R', 'G', 'B', 'A']
    present_rgba_names = []
    # Pulls out any present RGBA names in RGBA order.
    for name in conventional_rgba_names:
      if name in remaining_channel_names:
        present_rgba_names.append(name)
        remaining_channel_names.remove(name)
    channel_names = present_rgba_names + sorted(remaining_channel_names)

  return channels_to_ndarray(exr, channel_names), channel_names


def write_exr(filename, values, channel_names):
  """Writes the values in a multi-channel ndarray into an EXR file.

  Args:
    filename: The filename of the output file
    values: A numpy ndarray with shape [height, width, channels]
    channel_names: A list of strings with length = channels

  Raises:
    TypeError: If the numpy array has an unsupported type.
    ValueError: If the length of the array and the length of the channel names
      list do not match.
  """
  if values.shape[-1] != len(channel_names):
    raise ValueError(
        'Number of channels in values does not match channel names (%d, %d)' %
        (values.shape[-1], len(channel_names)))
  header = OpenEXR.Header(values.shape[1], values.shape[0])
  try:
    exr_channel_type = Imath.PixelType(_np_to_exr[values.dtype.type])
  except KeyError:
    raise TypeError('Unsupported numpy type: %s' % str(values.dtype))
  header['channels'] = {
      n: Imath.Channel(exr_channel_type) for n in channel_names
  }
  channel_data = [values[..., i] for i in range(values.shape[-1])]
  exr = OpenEXR.OutputFile(filename, header)
  exr.writePixels(
      dict((n, d.tobytes()) for n, d in zip(channel_names, channel_data)))
  exr.close()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Generates fake data for testing."""

import os
from absl import app
from absl import flags
import h5py
import numpy as np

flags.DEFINE_string("fakes_path", ".", "path where files will be generated")
FLAGS = flags.FLAGS


def main(argv):
  """Generates files with the internal structure.

  Args:
    argv: the path where to generate the fake files
  Reference: f = h5py.File("modelnet40_ply_hdf5_2048/ply_data_train0.h5", "r")
    print(f['data'])  # <HDF5 dataset "data": shape(2048, 2048, 3), type "<f4">
    print(f['label']) # <HDF5 dataset "label": shape(2048, 1), type "|u1">
  """
  if len(argv) != 1:
    raise app.UsageError("One argument required.")

  for i in range(3):
    fake_points = np.random.randn(8, 2048, 3).astype(np.float32)
    fake_label = np.random.uniform(low=0, high=40, size=(8, 1)).astype(np.uint8)
    path = os.path.join(FLAGS.fakes_path, "ply_data_train{}.h5".format(i))
    with h5py.File(path, "w") as h5f:
      h5f.create_dataset("data", data=fake_points)
      h5f.create_dataset("label", data=fake_label)
  for i in range(2):
    fake_points = np.random.randn(8, 2048, 3).astype(np.float32)
    fake_label = np.random.uniform(low=0, high=40, size=(8, 1)).astype(np.uint8)
    path = os.path.join(FLAGS.fakes_path, "ply_data_test{}.h5".format(i))
    with h5py.File(path, "w") as h5f:
      h5f.create_dataset("data", data=fake_points)
      h5f.create_dataset("label", data=fake_label)


if __name__ == "__main__":
  app.run(main)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""This module implements a Levenberg-Marquardt optimizer.

Minimizes \\(\min_{\mathbf{x}} \sum_i \|\mathbf{r}_i(\mathbf{x})\|^2_2\\) where
\\(\mathbf{r}_i(\mathbf{x})\\)
are the residuals. This function implements Levenberg-Marquardt, an iterative
process that linearizes the residuals and iteratively finds a displacement
\\(\Delta \mathbf{x}\\) such that at iteration \\(t\\) an update
\\(\mathbf{x}_{t+1} = \mathbf{x}_{t} + \Delta \mathbf{x}\\) improving the
loss can be computed. The displacement is computed by solving an optimization
problem
\\(\min_{\Delta \mathbf{x}} \sum_i
\|\mathbf{J}_i(\mathbf{x}_{t})\Delta\mathbf{x} +
\mathbf{r}_i(\mathbf{x}_t)\|^2_2 + \lambda\|\Delta \mathbf{x} \|_2^2\\) where
\\(\mathbf{J}_i(\mathbf{x}_{t})\\) is the Jacobian of \\(\mathbf{r}_i\\)
computed at \\(\mathbf{x}_t\\), and \\(\lambda\\) is a scalar weight.

More details on Levenberg-Marquardt can be found on [this page.]
(https://en.wikipedia.org/wiki/Levenberg%E2%80%93Marquardt_algorithm)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import export_api


def _values_and_jacobian(residuals, variables):
  """Computes the residual values and the Jacobian matrix.

  Args:
    residuals: A list of residuals.
    variables: A list of variables.

  Returns:
    The residual values and the Jacobian matrix.
  """

  def _compute_residual_values(residuals, variables):
    """Computes the residual values."""
    return tf.concat([
        tf.reshape(residual(*variables), shape=(-1,)) for residual in residuals
    ],
                     axis=-1)

  def _compute_jacobian(values, variables, tape):
    """Computes the Jacobian matrix."""
    jacobians = tape.jacobian(
        values, variables, unconnected_gradients=tf.UnconnectedGradients.ZERO)
    return tf.concat([
        tf.reshape(jacobian, shape=(tf.shape(input=jacobian)[0], -1))
        for jacobian in jacobians
    ],
                     axis=-1)

  with tf.GradientTape(watch_accessed_variables=False, persistent=True) as tape:
    for variable in variables:
      tape.watch(variable)
    values = _compute_residual_values(residuals, variables)
  jacobian = _compute_jacobian(values, variables, tape)
  del tape
  values = tf.expand_dims(values, axis=-1)
  return values, jacobian


def minimize(residuals,
             variables,
             max_iterations,
             regularizer=1e-20,
             regularizer_multiplier=10.0,
             callback=None,
             name=None):
  r"""Minimizes a set of residuals in the least-squares sense.

  Args:
    residuals: A residual or a list/tuple of residuals. A residual is a Python
      `callable`.
    variables: A variable or a list or tuple of variables defining the starting
      point of the minimization.
    max_iterations: The maximum number of iterations.
    regularizer: The regularizer is used to damped the stepsize when the
      iterations are becoming unstable. The bigger the regularizer is the
      smaller the stepsize becomes.
    regularizer_multiplier: If an iteration does not decrease the objective a
      new regularizer is computed by scaling it by this multiplier.
    callback: A callback function that will be called at each iteration. In
      graph mode the callback should return an op or list of ops that will
      execute the callback logic. The callback needs to be of the form
      f(iteration, objective_value, variables). A callback is a Python
      `callable`. The callback could be used for logging, for example if one
      wants to print the objective value at each iteration.
    name: A name for this op. Defaults to "levenberg_marquardt_minimize".

  Returns:
    The value of the objective function and variables attained at the final
    iteration of the minimization procedure.

  Raises:
    ValueError: If max_iterations is not at least 1.
    InvalidArgumentError: This exception is only raised in graph mode if the
    Cholesky decomposition is not successful. One likely fix is to increase
    the regularizer. In eager mode this exception is catched and the regularizer
    is increased automatically.

  Examples:

    ```python
    x = tf.constant(np.random.random_sample(size=(1,2)), dtype=tf.float32)
    y = tf.constant(np.random.random_sample(size=(3,1)), dtype=tf.float32)

    def f1(x, y):
      return x + y

    def f2(x, y):
      return x * y

    def callback(iteration, objective_value, variables):
      def print_output(iteration, objective_value, *variables):
        print("Iteration:", iteration, "Objective Value:", objective_value)
        for variable in variables:
          print(variable)
      inp = [iteration, objective_value] + variables
      return tf.py_function(print_output, inp, [])

    minimize_op = minimize(residuals=(f1, f2),
                           variables=(x, y),
                           max_iterations=10,
                           callback=callback)

    if not tf.executing_eagerly():
      with tf.Session() as sess:
        sess.run(tf.global_variables_initializer())
        sess.run(minimize_op)
    ```
  """
  if not isinstance(variables, (tuple, list)):
    variables = [variables]
  with tf.compat.v1.name_scope(name, 'levenberg_marquardt_minimize', variables):
    if not isinstance(residuals, (tuple, list)):
      residuals = [residuals]
    if isinstance(residuals, tuple):
      residuals = list(residuals)
    if isinstance(variables, tuple):
      variables = list(variables)
    variables = [tf.convert_to_tensor(value=variable) for variable in variables]
    multiplier = tf.constant(regularizer_multiplier, dtype=variables[0].dtype)

    if max_iterations <= 0:
      raise ValueError("'max_iterations' needs to be at least 1.")

    def _cond(iteration, regularizer, objective_value, variables):
      """Returns whether any iteration still needs to be performed."""
      del regularizer, objective_value, variables
      return iteration < max_iterations

    def _body(iteration, regularizer, objective_value, variables):
      """Main optimization loop."""
      iteration += tf.constant(1, dtype=tf.int32)
      values, jacobian = _values_and_jacobian(residuals, variables)
      # Solves the normal equation.
      try:
        updates = tf.linalg.lstsq(jacobian, values, l2_regularizer=regularizer)
        shapes = [tf.shape(input=variable) for variable in variables]
        splits = [tf.reduce_prod(input_tensor=shape) for shape in shapes]
        updates = tf.split(tf.squeeze(updates, axis=-1), splits)
        new_variables = [
            variable - tf.reshape(update, shape)
            for variable, update, shape in zip(variables, updates, shapes)
        ]
        new_objective_value = tf.reduce_sum(input_tensor=[
            tf.nn.l2_loss(residual(*new_variables)) for residual in residuals
        ])
        # If the new estimated solution does not decrease the objective value,
        # no updates are performed, but a new regularizer is computed.
        cond = tf.less(new_objective_value, objective_value)
        regularizer = tf.compat.v1.where(
            cond, x=regularizer, y=regularizer * multiplier)
        objective_value = tf.compat.v1.where(
            cond, x=new_objective_value, y=objective_value)
        variables = [
            tf.compat.v1.where(cond, x=new_variable, y=variable)
            for variable, new_variable in zip(variables, new_variables)
        ]
      # Note that catching InvalidArgumentError will only work in eager mode.
      except tf.errors.InvalidArgumentError:
        regularizer *= multiplier
      if callback is not None:
        callback_ops = callback(iteration, objective_value, variables)
        if callback_ops is not None:
          if not isinstance(callback_ops, (tuple, list)):
            callback_ops = [callback_ops]
          with tf.control_dependencies(callback_ops):
            iteration = tf.identity(iteration)
            objective_value = tf.identity(objective_value)
            variables = [tf.identity(v) for v in variables]
      return iteration, regularizer, objective_value, variables

    starting_value = tf.reduce_sum(input_tensor=[
        tf.nn.l2_loss(residual(*variables)) for residual in residuals
    ])
    dtype = variables[0].dtype
    initial = (
        tf.constant(0, dtype=tf.int32),  # Initial iteration number.
        tf.constant(regularizer, dtype=dtype),  # Initial regularizer.
        starting_value,  # Initial objective value.
        variables,  # Initial variables.
    )
    _, _, final_objective_value, final_variables = tf.while_loop(
        cond=_cond, body=_body, loop_vars=initial, parallel_iterations=1)
    return final_objective_value, final_variables


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for as_conformal_as_possible."""

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.deformation_energy import as_conformal_as_possible
from tensorflow_graphics.geometry.transformation import quaternion
from tensorflow_graphics.util import test_case


class AsConformalAsPossibleTest(test_case.TestCase):

  def test_energy_identity(self):
    """Checks that energy evaluated between the rest pose and itself is zero."""
    number_vertices = np.random.randint(3, 10)
    batch_size = np.random.randint(3)
    batch_shape = np.random.randint(1, 10, size=(batch_size)).tolist()
    vertices_rest_pose = np.random.uniform(size=(number_vertices, 3))
    vertices_deformed_pose = tf.broadcast_to(
        vertices_rest_pose, shape=batch_shape + [number_vertices, 3])
    quaternions = quaternion.from_euler(
        np.zeros(shape=batch_shape + [number_vertices, 3]))
    num_edges = int(round(number_vertices / 2))
    edges = np.zeros(shape=(num_edges, 2), dtype=np.int32)
    edges[..., 0] = np.linspace(
        0, number_vertices / 2 - 1, num_edges, dtype=np.int32)
    edges[..., 1] = np.linspace(
        number_vertices / 2, number_vertices - 1, num_edges, dtype=np.int32)

    energy = as_conformal_as_possible.energy(
        vertices_rest_pose=vertices_rest_pose,
        vertices_deformed_pose=vertices_deformed_pose,
        quaternions=quaternions,
        edges=edges,
        conformal_energy=False)

    self.assertAllClose(energy, tf.zeros_like(energy))

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_energy_jacobian_random(self):
    """Checks the correctness of the jacobian of energy."""
    number_vertices = np.random.randint(3, 10)
    batch_size = np.random.randint(3)
    batch_shape = np.random.randint(1, 10, size=(batch_size)).tolist()
    vertices_rest_pose_init = np.random.uniform(size=(number_vertices, 3))
    vertices_deformed_pose_init = np.random.uniform(size=batch_shape +
                                                    [number_vertices, 3])
    quaternions_init = np.random.uniform(size=batch_shape +
                                         [number_vertices, 4])
    num_edges = int(round(number_vertices / 2))
    edges = np.zeros(shape=(num_edges, 2), dtype=np.int32)
    edges[..., 0] = np.linspace(
        0, number_vertices / 2 - 1, num_edges, dtype=np.int32)
    edges[..., 1] = np.linspace(
        number_vertices / 2, number_vertices - 1, num_edges, dtype=np.int32)

    def conformal_energy(vertices_rest_pose, vertices_deformed_pose,
                         quaternions):
      return as_conformal_as_possible.energy(
          vertices_rest_pose=vertices_rest_pose,
          vertices_deformed_pose=vertices_deformed_pose,
          quaternions=quaternions,
          edges=edges,
          conformal_energy=True)

    def nonconformal_energy(vertices_rest_pose, vertices_deformed_pose,
                            quaternions):
      return as_conformal_as_possible.energy(
          vertices_rest_pose=vertices_rest_pose,
          vertices_deformed_pose=vertices_deformed_pose,
          quaternions=quaternions,
          edges=edges,
          conformal_energy=False)

    with self.subTest(name="conformal"):
      self.assert_jacobian_is_correct_fn(conformal_energy, [
          vertices_rest_pose_init, vertices_deformed_pose_init, quaternions_init
      ])

    with self.subTest(name="non_conformal"):
      self.assert_jacobian_is_correct_fn(nonconformal_energy, [
          vertices_rest_pose_init, vertices_deformed_pose_init, quaternions_init
      ])

  @parameterized.parameters(
      ((1, 3), (1, 3), (1, 4), (1, 2), (1,), (1,)),
      ((1, 3), (None, 1, 3), (None, 1, 4), (1, 2), (1,), (1,)),
      ((1, 3), (2, None, 1, 3), (2, None, 1, 4), (1, 2), (1,), (1,)),
  )
  def test_energy_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        as_conformal_as_possible.energy, shapes,
        [tf.float32, tf.float32, tf.float32, tf.int32, tf.float32, tf.float32])

  def test_energy_preset(self):
    """Checks that energy returns the expected value."""
    vertices_rest_pose = np.array(((1.0, 0.0, 0.0), (-1.0, 0.0, 0.0)))
    vertices_deformed_pose = 2.0 * vertices_rest_pose
    quaternions = quaternion.from_euler(
        np.zeros(shape=vertices_deformed_pose.shape))
    edges = ((0, 1),)

    all_weights_1_energy = as_conformal_as_possible.energy(
        vertices_rest_pose, vertices_deformed_pose, quaternions, edges)
    all_weights_1_gt = 4.0
    vertex_weights = np.array((2.0, 1.0))
    vertex_weights_energy = as_conformal_as_possible.energy(
        vertices_rest_pose=vertices_rest_pose,
        vertices_deformed_pose=vertices_deformed_pose,
        quaternions=quaternions,
        edges=edges,
        vertex_weight=vertex_weights)
    vertex_weights_gt = 10.0
    edge_weights = np.array((2.0,))
    edge_weights_energy = as_conformal_as_possible.energy(
        vertices_rest_pose=vertices_rest_pose,
        vertices_deformed_pose=vertices_deformed_pose,
        quaternions=quaternions,
        edges=edges,
        edge_weight=edge_weights)
    edge_weights_gt = 16.0

    with self.subTest(name="all_weights_1"):
      self.assertAllClose(all_weights_1_energy, all_weights_1_gt)
    with self.subTest(name="vertex_weights"):
      self.assertAllClose(vertex_weights_energy, vertex_weights_gt)
    with self.subTest(name="edge_weights"):
      self.assertAllClose(edge_weights_energy, edge_weights_gt)

  @parameterized.parameters(
      ("vertices_rest_pose must have exactly 3 dimensions in axis",
       (1, 2), (1, 3), (1, 4), (1, 2), (1,), (1,)),
      ("vertices_rest_pose must have a rank of 2",
       (2, 1, 2), (1, 3), (1, 4), (1, 2), (1,), (1,)),
      ("vertices_deformed_pose must have exactly 3 dimensions in axis",
       (1, 3), (1, 2), (1, 4), (1, 2), (1,), (1,)),
      ("must have the same number of dimensions",
       (1, 3), (2, 3), (1, 4), (1, 2), (1,), (1,)),
      ("quaternions must have exactly 4 dimensions in axis",
       (1, 3), (1, 3), (1, 5), (1, 2), (1,), (1,)),
      ("must have the same number of dimensions",
       (1, 3), (1, 3), (2, 4), (1, 2), (1,), (1,)),
      ("Not all batch dimensions are identical",
       (1, 3), (1, 3), (2, 1, 4), (1, 2), (1,), (1,)),
      ("edges must have exactly 2 dimensions in axis",
       (1, 3), (1, 3), (1, 4), (1, 3), (1,), (1,)),
      ("edges must have a rank of 2",
       (1, 3), (1, 3), (1, 4), (2, 1, 2), (1,), (1,)),
      ("must have the same number of dimensions",
       (1, 3), (1, 3), (1, 4), (1, 2), (2,), (1,)),
      ("must have the same number of dimensions",
       (1, 3), (1, 3), (1, 4), (1, 2), (1,), (2,)),
  )  # pyformat: disable
  def test_energy_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(
        as_conformal_as_possible.energy, error_msg, shapes,
        [tf.float32, tf.float32, tf.float32, tf.int32, tf.float32, tf.float32])


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""This module implements 2d rotation matrix functionalities.

Given an angle of rotation $$\theta$$ a 2d rotation matrix can be expressed as

$$
\mathbf{R} =
\begin{bmatrix}
\cos(\theta) & -\sin(\theta) \\
\sin(\theta) & \cos(\theta)
\end{bmatrix}.
$$

More details rotation matrices can be found on [this page.]
(https://en.wikipedia.org/wiki/Rotation_matrix)

Note: This matrix rotates points in the $$xy$$-plane counterclockwise.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.geometry.transformation import rotation_matrix_common
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def from_euler(angle, name=None):
  r"""Converts an angle to a 2d rotation matrix.

  Converts an angle $$\theta$$ to a 2d rotation matrix following the equation

  $$
  \mathbf{R} =
  \begin{bmatrix}
  \cos(\theta) & -\sin(\theta) \\
  \sin(\theta) & \cos(\theta)
  \end{bmatrix}.
  $$

  Note:
    The resulting matrix rotates points in the $$xy$$-plane counterclockwise.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    angle: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      represents an angle in radians.
    name: A name for this op that defaults to
      "rotation_matrix_2d_from_euler_angle".

  Returns:
    A tensor of shape `[A1, ..., An, 2, 2]`, where the last dimension represents
    a 2d rotation matrix.

  Raises:
    ValueError: If the shape of `angle` is not supported.
  """
  with tf.compat.v1.name_scope(name, "rotation_matrix_2d_from_euler_angle",
                               [angle]):
    angle = tf.convert_to_tensor(value=angle)

    shape.check_static(
        tensor=angle, tensor_name="angle", has_dim_equals=(-1, 1))

    cos_angle = tf.cos(angle)
    sin_angle = tf.sin(angle)
    matrix = tf.stack((cos_angle, -sin_angle,
                       sin_angle, cos_angle),
                      axis=-1)  # pyformat: disable
    output_shape = tf.concat((tf.shape(input=angle)[:-1], (2, 2)), axis=-1)
    return tf.reshape(matrix, shape=output_shape)


def from_euler_with_small_angles_approximation(angles, name=None):
  r"""Converts an angle to a 2d rotation matrix under the small angle assumption.

  Under the small angle assumption, $$\sin(x)$$ and $$\cos(x)$$ can be
  approximated by their second order Taylor expansions, where
  $$\sin(x) \approx x$$ and $$\cos(x) \approx 1 - \frac{x^2}{2}$$. The 2d
  rotation matrix will then be approximated as

  $$
  \mathbf{R} =
  \begin{bmatrix}
  1.0 - 0.5\theta^2 & -\theta \\
  \theta & 1.0 - 0.5\theta^2
  \end{bmatrix}.
  $$

   In the current implementation, the smallness of the angles is not verified.

  Note:
    The resulting matrix rotates points in the $$xy$$-plane counterclockwise.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    angles: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      represents a small angle in radians.
    name: A name for this op that defaults to
      "rotation_matrix_2d_from_euler_with_small_angles_approximation".

  Returns:
    A tensor of shape `[A1, ..., An, 2, 2]`, where the last dimension represents
    a 2d rotation matrix.

  Raises:
    ValueError: If the shape of `angle` is not supported.
  """
  with tf.compat.v1.name_scope(
      name, "rotation_matrix_2d_from_euler_with_small_angles_approximation",
      [angles]):
    angles = tf.convert_to_tensor(value=angles)

    shape.check_static(
        tensor=angles, tensor_name="angles", has_dim_equals=(-1, 1))

    cos_angle = 1.0 - 0.5 * angles * angles
    sin_angle = angles
    matrix = tf.stack((cos_angle, -sin_angle,
                       sin_angle, cos_angle),
                      axis=-1)  # pyformat: disable
    output_shape = tf.concat((tf.shape(input=angles)[:-1], (2, 2)), axis=-1)
    return tf.reshape(matrix, shape=output_shape)


def inverse(matrix, name=None):
  """Computes the inverse of a 2D rotation matrix.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    matrix: A tensor of shape `[A1, ..., An, 2, 2]`, where the last two
      dimensions represent a 2d rotation matrix.
    name: A name for this op that defaults to "rotation_matrix_2d_inverse".

  Returns:
    A tensor of shape `[A1, ..., An, 2, 2]`, where the last dimension represents
    a 2d rotation matrix.

  Raises:
    ValueError: If the shape of `matrix` is not supported.
  """
  with tf.compat.v1.name_scope(name, "rotation_matrix_2d_inverse", [matrix]):
    matrix = tf.convert_to_tensor(value=matrix)

    shape.check_static(
        tensor=matrix,
        tensor_name="matrix",
        has_rank_greater_than=1,
        has_dim_equals=((-2, 2), (-1, 2)))

    ndims = matrix.shape.ndims
    perm = list(range(ndims - 2)) + [ndims - 1, ndims - 2]
    return tf.transpose(a=matrix, perm=perm)


def is_valid(matrix, atol=1e-3, name=None):
  r"""Determines if a matrix is a valid rotation matrix.

  Determines if a matrix $$\mathbf{R}$$ is a valid rotation matrix by checking
  that $$\mathbf{R}^T\mathbf{R} = \mathbf{I}$$ and $$\det(\mathbf{R}) = 1$$.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    matrix: A tensor of shape `[A1, ..., An, 2, 2]`, where the last two
      dimensions represent a 2d rotation matrix.
    atol: The absolute tolerance parameter.
    name: A name for this op that defaults to "rotation_matrix_2d_is_valid".

  Returns:
    A tensor of type `bool` and shape `[A1, ..., An, 1]` where False indicates
    that the input is not a valid rotation matrix.
  """
  with tf.compat.v1.name_scope(name, "rotation_matrix_2d_is_valid", [matrix]):
    matrix = tf.convert_to_tensor(value=matrix)

    shape.check_static(
        tensor=matrix,
        tensor_name="matrix",
        has_rank_greater_than=1,
        has_dim_equals=((-2, 2), (-1, 2)))

    return rotation_matrix_common.is_valid(matrix, atol)


def rotate(point, matrix, name=None):
  """Rotates a 2d point using a 2d rotation matrix.

  Note:
    In the following, A1 to An are optional batch dimensions, which must be
    identical.

  Args:
    point: A tensor of shape `[A1, ..., An, 2]`, where the last dimension
      represents a 2d point.
    matrix: A tensor of shape `[A1, ..., An, 2, 2]`, where the last two
      dimensions represent a 2d rotation matrix.
    name: A name for this op that defaults to "rotation_matrix_2d_rotate".

  Returns:
    A tensor of shape `[A1, ..., An, 2]`, where the last dimension
      represents a 2d point.

  Raises:
    ValueError: If the shape of `point` or `matrix` is not supported.
  """
  with tf.compat.v1.name_scope(name, "rotation_matrix_2d_rotate",
                               [point, matrix]):
    point = tf.convert_to_tensor(value=point)
    matrix = tf.convert_to_tensor(value=matrix)

    shape.check_static(
        tensor=point, tensor_name="point", has_dim_equals=(-1, 2))
    shape.check_static(
        tensor=matrix,
        tensor_name="matrix",
        has_rank_greater_than=1,
        has_dim_equals=((-2, 2), (-1, 2)))
    shape.compare_batch_dimensions(
        tensors=(point, matrix),
        tensor_names=("point", "matrix"),
        last_axes=(-2, -3),
        broadcast_compatible=True)

    point = tf.expand_dims(point, axis=-1)
    common_batch_shape = shape.get_broadcasted_shape(
        point.shape[:-2], matrix.shape[:-2])
    def dim_value(dim):
      return 1 if dim is None else tf.compat.v1.dimension_value(dim)
    common_batch_shape = [dim_value(dim) for dim in common_batch_shape]
    point = tf.broadcast_to(point, common_batch_shape + [2, 1])
    matrix = tf.broadcast_to(matrix, common_batch_shape + [2, 2])
    rotated_point = tf.matmul(matrix, point)
    return tf.squeeze(rotated_point, axis=-1)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()
